from kerchunk import codecs

from importlib.metadata import version as _version

try:
    __version__ = _version("kerchunk")
except Exception:
    # Local copy or not installed with setuptools.
    # Disable minimum version checks on downstream libraries.
    __version__ = "9999"

__all__ = ["__version__"]


def set_reference_filesystem_cachable(cachable=True):
    """While experimenting with kerchunk and referenceFS, it can be convenient to not cache FS instances

    You may wish to call this function with ``False`` before any kerchunking session; leaving
    the instances cachable (the default) is what end-users will want, since it will be
    more efficient.
    """
    import fsspec

    fsspec.get_filesystem_class("reference").cachable = cachable



import io
import fsspec
import enum
import ujson

try:
    import tifffile
except ModuleNotFoundError:  # pragma: no cover
    raise ImportError(
        "tifffile is required for kerchunking TIFF files. Please install with "
        "`pip/conda install tifffile`."
    )

import kerchunk.utils


def tiff_to_zarr(
    urlpath,
    remote_options=None,
    target=None,
    target_options=None,
    inline_threshold=0,
    storage_options=None,
):
    """Wraps TIFFFile's fsspec writer to extract metadata as attributes

    Parameters
    ----------
    urlpath: str
        Location of input TIFF
    remote_options: dict
        pass these to fsspec when opening urlpath
    target: str
        Write JSON to this location. If not given, no file is output
    target_options: dict
        pass these to fsspec when opening target
    inline_threshold: int
        Bytes blocks smaller than this will be inlined
    storage_options: dict
        same as remote_options, for compatibility. If both are given, remote_options wins.

    Returns
    -------
    references dict
    """

    with fsspec.open(urlpath, **(remote_options or storage_options or {})) as of:
        url, name = urlpath.rsplit("/", 1)
        prot = of.fs.protocol
        remote_protocol = prot[0] if isinstance(prot, tuple) else prot

        with tifffile.TiffFile(of, name=name) as tif:
            with tif.series[0].aszarr() as store:
                of2 = io.StringIO()
                store.write_fsspec(of2, url=url)
                out = ujson.loads(of2.getvalue())

                meta = ujson.loads(out[".zattrs"])
                for k in dir(tif):
                    if not k.endswith("metadata"):
                        continue
                    try:
                        met = getattr(tif, k, None)
                    except Exception:
                        continue
                    try:
                        d = dict(met or {})
                    except ValueError:
                        # newer tifffile exposes xml structured tags
                        from xml.etree import ElementTree

                        e = ElementTree.fromstring(met)
                        d = {i.get("name"): i.text for i in e}
                    meta.update(d)
                for k, v in meta.copy().items():
                    # deref enums
                    if isinstance(v, enum.EnumMeta):
                        meta[k] = v._name_
                out[".zattrs"] = ujson.dumps(meta)
    if "GTRasterTypeGeoKey" in meta:
        # TODO: make dataset and assign coords for geoTIFF
        # import zarr
        # fs = fsspec.filesystem("reference", fo=out
        # z = zarr.open(out.get_mapper())
        # coords = generate_coords(meta, z.shape)
        # rasterio.crs.CRS.from_epsg(attrs['ProjectedCSTypeGeoKey']).to_wkt("WKT1_GDAL") ??
        pass
    if inline_threshold:
        out = kerchunk.utils.do_inline(
            out, inline_threshold, remote_options, remote_protocol
        )
    if target is not None:
        with fsspec.open(target, **(target_options or {})) as of:
            ujson.dump(out, of)
    return out


# http://geotiff.maptools.org/spec/geotiff6.html#6.3.1.3
units = {
    9001: "metre",
    9002: "foot",
    9003: "US survey foot",
    9015: "mile international nautical",  # ... and many more
}


TiffToZarr = kerchunk.utils.class_factory(tiff_to_zarr)


def generate_coords(attrs, shape):
    """Produce coordinate arrays for given variable

    Specific to GeoTIFF input attributes

    Parameters
    ----------
    attrs: dict
        Containing the geoTIFF tags, probably the root group of the dataset
    shape: tuple[int]
        The array size in numpy (C) order
    """
    import numpy as np

    height, width = shape[-2:]
    xscale, yscale, zscale = attrs["ModelPixelScale"][:3]
    x0, y0, z0 = attrs["ModelTiepoint"][3:6]
    out = {}
    out["x"] = np.arange(width) * xscale + x0 + xscale / 2
    out["y"] = np.arange(height) * -yscale + y0 - yscale / 2
    if len(shape) > 2:
        out["z"] = np.arange(shape[-3]) * zscale + z0 + zscale / 2
    return out



import io
import logging
import pathlib
from contextlib import ExitStack
from typing import Union, Any, Dict, List, Tuple

import fsspec.core
from fsspec.implementations.reference import LazyReferenceMapper
import numpy as np
import zarr
import numcodecs
from numcodecs.abc import Codec
from .codecs import FillStringsCodec
from .utils import (
    _encode_for_JSON,
    encode_fill_value,
    dict_to_store,
    translate_refs_serializable,
)

try:
    import h5py
except ModuleNotFoundError:  # pragma: no cover
    raise ImportError(
        "h5py is required for kerchunking HDF5/NetCDF4 files. Please install with "
        "`pip/conda install h5py`. See https://docs.h5py.org/en/latest/build.html "
        "for more details."
    )

try:
    import hdf5plugin  # noqa
except ModuleNotFoundError:
    hdf5plugin = None

lggr = logging.getLogger("h5-to-zarr")
_HIDDEN_ATTRS = {  # from h5netcdf.attrs
    "REFERENCE_LIST",
    "CLASS",
    "DIMENSION_LIST",
    "NAME",
    "_Netcdf4Dimid",
    "_Netcdf4Coordinates",
    "_nc3_strict",
    "_NCProperties",
}


class Hdf5FeatureNotSupported(RuntimeError):
    pass


class SingleHdf5ToZarr:
    """Translate the content of one HDF5 file into Zarr metadata.

    HDF5 groups become Zarr groups. HDF5 datasets become Zarr arrays. Zarr array
    chunks remain in the HDF5 file.

    It is a good idea to call `.close()` on this instance after use, to free up
    underlying resources in the file and hdf objects.

    Parameters
    ----------
    h5f : file-like or str
        Input HDF5 file. Can be a binary Python file-like object (duck-typed, adhering
        to BinaryIO is optional), in which case must also provide url. If a str,
        file will be opened using fsspec and storage_options.
    url : string
        URI of the HDF5 file, if passing a file-like object or h5py File/Group
    spec : int
        The version of output to produce (see README of this repo)
    inline_threshold : int
        Include chunks smaller than this value directly in the output. Zero or negative
        to disable
    storage_options: dict
        passed to fsspec if h5f is a str
    error: "warn" (default) | "pdb" | "ignore" | "raise"
    vlen_encode: ["embed", "null", "leave", "encode"]
        What to do with VLEN string variables or columns of tabular variables
        leave: pass through the 16byte garbage IDs unaffected, but requires no codec
        null: set all the strings to None or empty; required that this library is available
        at read time
        embed: include all the values in the output JSON (should not be used for large tables)
        encode: save the ID-to-value mapping in a codec, to produce the real values at read
        time; requires this library to be available. Can be efficient storage where there
        are few unique values.
    out: dict-like or None
        This allows you to supply an fsspec.implementations.reference.LazyReferenceMapper
        to write out parquet as the references get filled, or some other dictionary-like class
        to customise how references get stored
    unsupported_inline_threshold: int or None
        Include chunks involving unsupported h5 features and smaller than this value directly
        in the output. None will use "inline_threshold".
    """

    def __init__(
        self,
        h5f: "io.BinaryIO | str | h5py.File | h5py.Group",
        url: str = None,
        spec=1,
        inline_threshold=500,
        storage_options=None,
        error="warn",
        vlen_encode="embed",
        out=None,
        unsupported_inline_threshold=None,
    ):

        # Open HDF5 file in read mode...
        lggr.debug(f"HDF5 file: {h5f}")
        self._closers = ExitStack()
        if isinstance(h5f, (pathlib.Path, str)):
            fs, path = fsspec.core.url_to_fs(h5f, **(storage_options or {}))
            self.input_file = self._closers.enter_context(fs.open(path, "rb"))
            url = h5f
            self._h5f = self._closers.enter_context(
                h5py.File(self.input_file, mode="r")
            )
        elif isinstance(h5f, io.IOBase):
            self.input_file = h5f
            self._h5f = self._closers.enter_context(
                h5py.File(self.input_file, mode="r")
            )
        elif isinstance(h5f, (h5py.File, h5py.Group)):
            # assume h5py object (File or group/dataset)
            self._h5f = h5f
            fs, path = fsspec.core.url_to_fs(url, **(storage_options or {}))
            self.input_file = self._closers.enter_context(fs.open(path, "rb"))
        else:
            raise ValueError("type of input `h5f` not recognised")
        self.spec = spec
        self.inline = inline_threshold
        if vlen_encode not in ["embed", "null", "leave", "encode"]:
            raise NotImplementedError
        self.vlen = vlen_encode
        self.store_dict = out or {}
        self.store = dict_to_store(self.store_dict)
        self._zroot = zarr.group(store=self.store, zarr_format=2)
        self._uri = url
        self.error = error
        if unsupported_inline_threshold is None:
            unsupported_inline_threshold = inline_threshold or 100
        self.unsupported_inline_threshold = unsupported_inline_threshold
        lggr.debug(f"HDF5 file URI: {self._uri}")

    def close(self):
        self._closers.close()

    def translate(self, preserve_linked_dsets=False):
        """Translate content of one HDF5 file into Zarr storage format.

        This method is the main entry point to execute the workflow, and
        returns a "reference" structure to be used with zarr/kerchunk

        No data is copied out of the HDF5 file.

        Parameters
        ----------
        preserve_linked_dsets : bool (optional, default False)
            If True, translate HDF5 soft and hard links for each `h5py.Dataset`
            into the reference structure. Requires h5py version 3.11.0 or later.
            Will not translate external links or links to `h5py.Group` objects.

        Returns
        -------
        dict
            Dictionary containing reference structure.
        """
        lggr.debug("Translation begins")
        self._transfer_attrs(self._h5f, self._zroot)
        self._h5f.visititems(self._translator)

        if preserve_linked_dsets:
            if not has_visititems_links():
                raise Hdf5FeatureNotSupported(
                    "'preserve_linked_dsets' kwarg requires h5py 3.11.0 or later "
                    f"is installed, found {h5py.__version__}"
                )
            self._h5f.visititems_links(self._translator)

        if self.spec < 1:
            return self.store
        elif isinstance(self.store, LazyReferenceMapper):
            self.store.flush()
            return self.store
        else:
            translate_refs_serializable(self.store_dict)
            store = _encode_for_JSON(self.store_dict)
            return {"version": 1, "refs": store}

    def _unref(self, ref):
        name = h5py.h5r.get_name(ref, self._h5f.id)
        return self._h5f[name]

    def _transfer_attrs(
        self,
        h5obj: Union[h5py.Dataset, h5py.Group],
        zobj: Union[zarr.Array, zarr.Group],
    ):
        """Transfer attributes from an HDF5 object to its equivalent Zarr object.

        Parameters
        ----------
        h5obj : h5py.Group or h5py.Dataset
            An HDF5 group or dataset.
        zobj : zarr.hierarchy.Group or zarr.core.Array
            An equivalent Zarr group or array to the HDF5 group or dataset with
            attributes.
        """
        upd = {}
        for n, v in h5obj.attrs.items():
            if n in _HIDDEN_ATTRS:
                continue

            # Fix some attribute values to avoid JSON encoding exceptions...
            if isinstance(v, bytes):
                v = v.decode("utf-8", errors="replace") or " "
            elif isinstance(v, (np.ndarray, np.number, np.bool_)):
                if v.dtype.kind == "S":
                    v = v.astype(str)
                if n == "_FillValue":
                    continue  # strip it out!
                elif v.size == 1:
                    v = v.flatten()[0]
                    if isinstance(v, (np.ndarray, np.number, np.bool_)):
                        v = v.tolist()
                else:
                    v = v.tolist()
            elif isinstance(v, h5py._hl.base.Empty):
                v = ""
            if v == "DIMENSION_SCALE":
                continue
            try:
                if isinstance(v, (str, int, float)):
                    upd[n] = v
                elif isinstance(v, (tuple, set, list)) and all(
                    isinstance(_, (str, int, float)) for _ in v
                ):
                    upd[n] = list(v)
                else:
                    upd[n] = str(v)
            except TypeError:
                lggr.debug(
                    f"TypeError transferring attr, skipping:\n {n}@{h5obj.name} = {v} ({type(v)})"
                )
        zobj.attrs.update(upd)

    def _decode_filters(self, h5obj: Union[h5py.Dataset, h5py.Group]):
        if h5obj.scaleoffset:
            raise Hdf5FeatureNotSupported(
                f"{h5obj.name} uses HDF5 scaleoffset filter - not supported by kerchunk"
            )
        if h5obj.compression in ("szip", "lzf"):
            raise Hdf5FeatureNotSupported(
                f"{h5obj.name} uses szip or lzf compression - not supported by kerchunk"
            )
        filters = []
        if h5obj.shuffle and h5obj.dtype.kind != "O":
            # cannot use shuffle if we materialised objects
            filters.append(numcodecs.Shuffle(elementsize=h5obj.dtype.itemsize))
        for filter_id, properties in h5obj._filters.items():
            if str(filter_id) == "32001":
                blosc_compressors = (
                    "blosclz",
                    "lz4",
                    "lz4hc",
                    "snappy",
                    "zlib",
                    "zstd",
                )
                (
                    _1,
                    _2,
                    bytes_per_num,
                    total_bytes,
                    clevel,
                    shuffle,
                    compressor,
                ) = properties
                pars = dict(
                    blocksize=total_bytes,
                    clevel=clevel,
                    shuffle=shuffle,
                    cname=blosc_compressors[compressor],
                )
                filters.append(numcodecs.Blosc(**pars))
            elif str(filter_id) == "32015":
                filters.append(numcodecs.Zstd(level=properties[0]))
            elif str(filter_id) == "gzip":
                filters.append(numcodecs.Zlib(level=properties))
            elif str(filter_id) == "32004":
                raise Hdf5FeatureNotSupported(
                    f"{h5obj.name} uses lz4 compression - not supported by kerchunk"
                )
            elif str(filter_id) == "32008":
                raise Hdf5FeatureNotSupported(
                    f"{h5obj.name} uses bitshuffle compression - not supported by kerchunk"
                )
            elif str(filter_id) == "shuffle":
                # already handled before this loop
                pass
            else:
                raise Hdf5FeatureNotSupported(
                    f"{h5obj.name} uses filter id {filter_id} with properties {properties},"
                    f" not supported by kerchunk."
                )
        return filters

    def _translator(
        self,
        name: str,
        h5obj: Union[
            h5py.Dataset, h5py.Group, h5py.SoftLink, h5py.HardLink, h5py.ExternalLink
        ],
    ):
        """Produce Zarr metadata for all groups and datasets in the HDF5 file."""
        try:  # method must not raise exception
            kwargs: Dict[str, Any] = {"compressor": None}

            if isinstance(h5obj, (h5py.SoftLink, h5py.HardLink)):
                h5obj = self._h5f[name]
                if isinstance(h5obj, h5py.Group):
                    # continues iteration of visititems_links
                    lggr.debug(
                        f"Skipping translation of HDF5 linked group: '{h5obj.name}'"
                    )
                    return None

            if isinstance(h5obj, h5py.Dataset):
                lggr.debug(f"HDF5 dataset: {h5obj.name}")
                lggr.debug(f"HDF5 compression: {h5obj.compression}")
                if h5obj.id.get_create_plist().get_layout() == h5py.h5d.COMPACT:
                    # Only do if h5obj.nbytes < self.inline??
                    # kwargs["data"] = h5obj[:]
                    if h5obj.nbytes < self.inline:
                        kwargs["data"] = h5obj[:]
                    kwargs["filters"] = []
                else:
                    try:
                        kwargs["filters"] = self._decode_filters(h5obj)
                    except Hdf5FeatureNotSupported:
                        if h5obj.nbytes < self.unsupported_inline_threshold:
                            kwargs["data"] = _read_unsupported_direct(h5obj)
                            kwargs["filters"] = []
                        else:
                            raise
                dt = None
                # Get storage info of this HDF5 dataset...
                if "data" in kwargs:
                    cinfo = _NULL_CHUNK_INFO
                else:
                    try:
                        cinfo = self._storage_info_and_adj_filters(
                            h5obj, kwargs["filters"]
                        )
                    except Hdf5FeatureNotSupported:
                        if h5obj.nbytes < self.unsupported_inline_threshold:
                            kwargs["data"] = _read_unsupported_direct(h5obj)
                            kwargs["filters"] = []
                            cinfo = _NULL_CHUNK_INFO
                        else:
                            raise

                if "data" in kwargs:
                    fill = None
                else:
                    # encodings
                    if h5obj.dtype.kind in "US":
                        fill = h5obj.fillvalue or " "  # cannot be None
                    elif h5obj.dtype.kind == "O":
                        if self.vlen == "embed":
                            if np.isscalar(h5obj):
                                out = str(h5obj)
                            elif h5obj.ndim == 0:
                                out = np.array(h5obj).tolist().decode()
                            else:
                                out = h5obj[:]
                                out2 = out.ravel()
                                for i, val in enumerate(out2):
                                    if isinstance(val, bytes):
                                        out2[i] = val.decode()
                                    elif isinstance(val, str):
                                        out2[i] = val
                                    elif isinstance(val, h5py.h5r.Reference):
                                        # TODO: recursively recreate references
                                        out2[i] = None
                                    else:
                                        out2[i] = [
                                            v.decode() if isinstance(v, bytes) else v
                                            for v in val
                                        ]
                            kwargs["data"] = out
                            kwargs["filters"] = [numcodecs.VLenUTF8()]
                            fill = None
                        elif self.vlen == "null":
                            dt = "O"
                            kwargs["filters"] = [FillStringsCodec(dtype="S16")]
                            fill = " "
                        elif self.vlen == "leave":
                            dt = "S16"
                            fill = " "
                        elif self.vlen == "encode":
                            assert len(cinfo) == 1
                            v = list(cinfo.values())[0]
                            data = _read_block(self.input_file, v["offset"], v["size"])
                            indexes = np.frombuffer(data, dtype="S16")
                            labels = h5obj[:]
                            mapping = {
                                index.decode(): label.decode()
                                for index, label in zip(indexes, labels)
                            }
                            kwargs["filters"] = [
                                FillStringsCodec(dtype="S16", id_map=mapping)
                            ]
                            fill = " "
                        else:
                            raise NotImplementedError
                    elif _is_netcdf_datetime(h5obj) or _is_netcdf_variable(h5obj):
                        fill = None
                    else:
                        fill = h5obj.fillvalue
                    if h5obj.dtype.kind == "V":
                        fill = None
                        if self.vlen == "encode":
                            assert len(cinfo) == 1
                            v = list(cinfo.values())[0]
                            dt = [
                                (
                                    v,
                                    (
                                        "S16"
                                        if h5obj.dtype[v].kind == "O"
                                        else str(h5obj.dtype[v])
                                    ),
                                )
                                for v in h5obj.dtype.names
                            ]
                            data = _read_block(self.input_file, v["offset"], v["size"])
                            labels = h5obj[:]
                            arr = np.frombuffer(data, dtype=dt)
                            mapping = {}
                            for field in labels.dtype.names:
                                if labels[field].dtype == "O":
                                    mapping.update(
                                        {
                                            index.decode(): label.decode()
                                            for index, label in zip(
                                                arr[field], labels[field]
                                            )
                                        }
                                    )
                            kwargs["filters"] = [
                                FillStringsCodec(dtype=str(dt), id_map=mapping)
                            ]
                            dt = [
                                (
                                    v,
                                    (
                                        "O"
                                        if h5obj.dtype[v].kind == "O"
                                        else str(h5obj.dtype[v])
                                    ),
                                )
                                for v in h5obj.dtype.names
                            ]
                        elif self.vlen == "null":
                            dt = [
                                (
                                    v,
                                    (
                                        "S16"
                                        if h5obj.dtype[v].kind == "O"
                                        else str(h5obj.dtype[v])
                                    ),
                                )
                                for v in h5obj.dtype.names
                            ]
                            kwargs["filters"] = [FillStringsCodec(dtype=str(dt))]
                            dt = [
                                (
                                    v,
                                    (
                                        "O"
                                        if h5obj.dtype[v].kind == "O"
                                        else str(h5obj.dtype[v])
                                    ),
                                )
                                for v in h5obj.dtype.names
                            ]
                        elif self.vlen == "leave":
                            dt = [
                                (
                                    v,
                                    (
                                        "S16"
                                        if h5obj.dtype[v].kind == "O"
                                        else h5obj.dtype[v]
                                    ),
                                )
                                for v in h5obj.dtype.names
                            ]
                        elif self.vlen == "embed":
                            # embed fails due to https://github.com/zarr-developers/numcodecs/issues/333
                            data = h5obj[:].tolist()
                            data2 = []
                            for d in data:
                                data2.append(
                                    [
                                        (
                                            _.decode(errors="ignore")
                                            if isinstance(_, bytes)
                                            else _
                                        )
                                        for _ in d
                                    ]
                                )
                            dt = "O"
                            kwargs["data"] = data2
                            kwargs["filters"] = [numcodecs.VLenUTF8()]
                            fill = None
                        else:
                            raise NotImplementedError

                    if h5py.h5ds.is_scale(h5obj.id) and not cinfo:
                        return
                    if h5obj.attrs.get("_FillValue") is not None:
                        fill = h5obj.attrs.get("_FillValue")
                        fill = encode_fill_value(fill, dt or h5obj.dtype)

                adims = self._get_array_dims(h5obj)

                # Create a Zarr array equivalent to this HDF5 dataset.
                data = kwargs.pop("data", None)
                if (dt or h5obj.dtype) == object:
                    dt = "string"
                za = self._zroot.require_array(
                    name=h5obj.name.lstrip("/"),
                    shape=h5obj.shape,
                    dtype=dt or h5obj.dtype,
                    chunks=h5obj.chunks or h5obj.shape,
                    fill_value=fill,
                    attributes={
                        "_ARRAY_DIMENSIONS": adims,
                    },
                    overwrite=True,
                    **kwargs,
                )
                lggr.debug(f"Created Zarr array: {za}")
                self._transfer_attrs(h5obj, za)

                lggr.debug(f"_ARRAY_DIMENSIONS = {adims}")
                if data is not None:
                    try:
                        za[:] = data
                    except (ValueError, TypeError):
                        store_key = f"{za.path}/{'.'.join('0' * h5obj.ndim)}"
                        self.store_dict[store_key] = _filters_encode_data(
                            data, kwargs["filters"]
                        )
                    return

                # Store chunk location metadata...
                if cinfo:
                    for k, v in cinfo.items():
                        key = (
                            str.removeprefix(h5obj.name, "/")
                            + "/"
                            + ".".join(map(str, k))
                        )
                        if "data" in v:
                            self.store_dict[key] = v["data"]
                            continue
                        if h5obj.fletcher32:
                            logging.info("Discarding fletcher32 checksum")
                            v["size"] -= 4
                        key = (
                            str.removeprefix(h5obj.name, "/")
                            + "/"
                            + ".".join(map(str, k))
                        )

                        if (
                            self.inline
                            and isinstance(v, dict)
                            and v["size"] < self.inline
                        ):
                            self.input_file.seek(v["offset"])
                            data = self.input_file.read(v["size"])
                            # Removed the encoding, as will finally be encoded_for_JSON
                            # try:
                            #     # easiest way to test if data is ascii
                            #     data.decode("ascii")
                            # except UnicodeDecodeError:
                            #     data = b"base64:" + base64.b64encode(data)

                            self.store_dict[key] = data
                        else:
                            self.store_dict[key] = [
                                self._uri,
                                v["offset"],
                                v["size"],
                            ]

            elif isinstance(h5obj, h5py.Group):
                lggr.debug(f"HDF5 group: {h5obj.name}")
                zgrp = self._zroot.require_group(h5obj.name.lstrip("/"))
                self._transfer_attrs(h5obj, zgrp)
        except Exception as e:
            import traceback

            msg = "\n".join(
                [
                    "The following exception was caught and quashed while traversing HDF5",
                    str(e),
                    traceback.format_exc(limit=5),
                ]
            )
            if self.error == "ignore":
                return
            elif self.error == "pdb":
                print(msg)
                import pdb

                pdb.post_mortem()
            elif self.error == "raise":
                raise
            else:
                # "warn" or anything else, the default
                import warnings

                warnings.warn(msg)
            del e  # garbage collect

    def _get_array_dims(self, dset):
        """Get a list of dimension scale names attached to input HDF5 dataset.

        This is required by the xarray package to work with Zarr arrays. Only
        one dimension scale per dataset dimension is allowed. If dataset is
        dimension scale, it will be considered as the dimension to itself.

        Parameters
        ----------
        dset : h5py.Dataset
            HDF5 dataset.

        Returns
        -------
        list
            List with HDF5 path names of dimension scales attached to input
            dataset.
        """
        dims = list()
        rank = len(dset.shape)
        if rank:
            for n in range(rank):
                num_scales = len(dset.dims[n])
                if num_scales == 1:
                    dims.append(dset.dims[n][0].name[1:])
                elif h5py.h5ds.is_scale(dset.id):
                    dims.append(dset.name[1:])
                elif num_scales > 1:
                    raise Hdf5FeatureNotSupported(
                        f"{dset.name}: {len(dset.dims[n])} "
                        f"dimension scales attached to dimension #{n}"
                    )
                elif num_scales == 0:
                    # Some HDF5 files do not have dimension scales.
                    # If this is the case, `num_scales` will be 0.
                    # In this case, we mimic netCDF4 and assign phony dimension names.
                    # See https://github.com/fsspec/kerchunk/issues/41

                    dims.append(f"phony_dim_{n}")
        return dims

    def _storage_info_and_adj_filters(self, dset: h5py.Dataset, filters: list) -> dict:
        """Get storage information of an HDF5 dataset in the HDF5 file.

        Storage information consists of file offset and size (length) for every
        chunk of the HDF5 dataset. HDF5 dataset also configs for each chunk
        which filters are skipped by `filter_mask` (mostly in the case where a chunk is small
        or when write directly with low-level api like `H5Dwrite_chunk`/`DatasetID.write_direct_chunk`),
        hence a filter will be cleared if the first chunk does not apply it.
        `Hdf5FeatherNotSupported` will be raised if chucks have heterogeneous `filter_mask` and alien chunks
        cannot be inlined.

        Parameters
        ----------
        dset : h5py.Dataset
            HDF5 dataset for which to collect storage information.
        filters: list
            List of filters to apply to the HDF5 dataset. Will be modified in place
            if some filters not applied.

        Returns
        -------
        dict
            HDF5 dataset storage information. Dict keys are chunk array offsets
            as tuples. Dict values are pairs with chunk file offset and size
            integers, or data content if in need of inline for some reason.
        """
        # Empty (null) dataset...
        if dset.shape is None:
            return _NULL_CHUNK_INFO

        dsid = dset.id
        if dset.chunks is None:
            # Contiguous dataset...
            if dsid.get_offset() is None:
                # No data ever written...
                return _NULL_CHUNK_INFO
            else:
                key = (0,) * (len(dset.shape) or 1)
                return {
                    key: {"offset": dsid.get_offset(), "size": dsid.get_storage_size()}
                }
        else:
            # Chunked dataset...
            num_chunks = dsid.get_num_chunks()
            if num_chunks == 0:
                # No data ever written...
                return _NULL_CHUNK_INFO

            # Go over all the dataset chunks...
            stinfo = dict()
            chunk_size = dset.chunks
            filter_mask = None  # type: None | int

            def get_key(blob):
                return tuple([a // b for a, b in zip(blob.chunk_offset, chunk_size)])

            def filter_filters_by_mask(
                filter_list: List[Codec], filter_mask_
            ) -> List[Codec]:
                # use [2:] to remove the heading `0b` and [::-1] to reverse the order
                bin_mask = bin(filter_mask_)[2:][::-1]
                filters_rest = [
                    ifilter
                    for ifilter, imask in zip(filter_list, bin_mask)
                    if imask == "0"
                ]
                return filters_rest

            def store_chunk_info(blob):
                nonlocal filter_mask
                if filter_mask is None:
                    filter_mask = blob.filter_mask
                elif filter_mask != blob.filter_mask:
                    if blob.size < self.unsupported_inline_threshold:
                        data_slc = tuple(
                            slice(dim_offset, min(dim_offset + dim_chunk, dim_bound))
                            for dim_offset, dim_chunk, dim_bound in zip(
                                blob.chunk_offset, chunk_size, dset.shape
                            )
                        )
                        data = _read_unsupported_direct(dset, data_slc)
                        if data.shape != chunk_size:
                            bg = np.full(
                                chunk_size, dset.fillvalue, dset.dtype, order="C"
                            )
                            bg[tuple(slice(0, d) for d in data.shape)] = data
                            data_flatten = bg.reshape(-1)
                        else:
                            data_flatten = data.reshape(-1)
                        encoded = _filters_encode_data(
                            data_flatten, filter_filters_by_mask(filters, filter_mask)
                        )
                        stinfo[get_key(blob)] = {"data": encoded}
                        return
                    else:
                        raise Hdf5FeatureNotSupported(
                            f"Dataset {dset.name} has heterogeneous `filter_mask` - "
                            f"not supported by kerchunk."
                        )
                stinfo[get_key(blob)] = {"offset": blob.byte_offset, "size": blob.size}

            has_chunk_iter = callable(getattr(dsid, "chunk_iter", None))

            if has_chunk_iter:
                dsid.chunk_iter(store_chunk_info)
            else:
                for index in range(num_chunks):
                    store_chunk_info(dsid.get_chunk_info(index))

            # In most cases, the `filter_mask` should be zero, which means that all filters are applied.
            # If a filter is skipped, the corresponding bit in the mask fill be set 1.
            if filter_mask is not None and filter_mask != 0:
                filters[:] = filter_filters_by_mask(filters, filter_mask)
            return stinfo


def _read_unsupported_direct(dset, slc: Union[slice, Tuple[slice, ...]] = slice(None)):
    try:
        return dset[slc]
    # As what I googled, you might get OSError/ValueError/RuntimeError
    # when a filter in need was not registered. I met an `OSError` with
    # bizarre error message in my test without `hdf5plugin` imported.
    # Just simply catch all exceptions, as we will rethrow it anyway.
    except Exception:
        if hdf5plugin is None:
            import warnings

            warnings.warn(
                "Attempt to directly read h5-dataset via `h5py` failed. It is recommended to "
                "install `hdf5plugin` so that we can register extra filters, and then try again."
            )
        raise


def _simple_type(x):
    if isinstance(x, bytes):
        return x.decode()
    if isinstance(x, np.number):
        if x.dtype.kind == "i":
            return int(x)
        return float(x)
    return x


def _read_block(open_file, offset, size):
    place = open_file.tell()
    open_file.seek(offset)
    data = open_file.read(size)
    open_file.seek(place)
    return data


def _is_netcdf_datetime(dataset: h5py.Dataset):
    units = dataset.attrs.get("units")
    if isinstance(units, bytes):
        units = units.decode("utf-8")
    # This is the same heuristic used by xarray
    # https://github.com/pydata/xarray/blob/f8bae5974ee2c3f67346298da12621af4cae8cf8/xarray/coding/times.py#L670
    return units and "since" in units


def _is_netcdf_variable(dataset: h5py.Dataset):
    return any("_Netcdf4" in _ for _ in dataset.attrs)


def has_visititems_links():
    return hasattr(h5py.Group, "visititems_links")


def _filters_encode_data(data, filters: List[Codec]):
    for ifilter in filters:
        data = ifilter.encode(data)
    return bytes(data)


_NULL_CHUNK_INFO = dict()



import glob
import os.path
import zarr
import pytest
import xarray as xr

from kerchunk.utils import refs_as_store

pytest.importorskip("tifffile")
pytest.importorskip("rioxarray")
import kerchunk.tiff

files = glob.glob(os.path.join(os.path.dirname(__file__), "lc*tif"))


def test_one():
    fn = files[0]
    out = kerchunk.tiff.tiff_to_zarr(fn)
    store = refs_as_store(out)
    z = zarr.open(store, zarr_format=2)
    assert list(z) == ["0", "1", "2"]
    assert z.attrs["multiscales"] == [
        {
            "datasets": [{"path": "0"}, {"path": "1"}, {"path": "2"}],
            "metadata": {},
            "name": "",
            "version": "0.1",
        }
    ]
    assert z["0"].shape == (2048, 2048)
    assert z["0"][:].max() == 8


def test_one_class():
    from collections import Counter

    fn = files[0]
    out = kerchunk.tiff.TiffToZarr(fn, inline_threshold=1000).translate()

    # test inlineing
    c = Counter(type(_) for k, _ in out.items() if ".z" not in k)
    assert c[list] == 72
    assert c[bytes] == 24

    # test zarr output
    store = refs_as_store(out)
    z = zarr.open(store, zarr_format=2)
    assert list(z) == ["0", "1", "2"]
    assert z.attrs["multiscales"] == [
        {
            "datasets": [{"path": "0"}, {"path": "1"}, {"path": "2"}],
            "metadata": {},
            "name": "",
            "version": "0.1",
        }
    ]
    assert z["0"].shape == (2048, 2048)
    assert z["0"][:].max() == 8


def test_coord():
    fn = files[0]
    out = kerchunk.tiff.tiff_to_zarr(fn)
    store = refs_as_store(out)
    z = zarr.open(store, zarr_format=2)  # highest res is the one xarray picks
    out = kerchunk.tiff.generate_coords(z.attrs, z["0"].shape)

    ds = xr.open_dataset(fn)
    assert (ds.x == out["x"]).all()
    assert (ds.y == out["y"]).all()



import base64
import logging

import ujson
import pandas as pd
import fsspec
import zarr


# example from preffs's README'
df = pd.DataFrame(
    {
        "key": ["a/b", "a/b", "b"],
        "path": ["a.dat", "b.dat", None],
        "offset": [123, 12, 0],
        "size": [12, 17, 0],
        "raw": [None, None, b"data"],
    }
)

logger = logging.getLogger("kerchunk.df")


def _proc_raw(r):
    if hasattr(r, "to_bytes"):
        r = r.to_bytes()
    elif not isinstance(r, bytes):
        r = r.encode()
    if r.startswith(b"base64:"):
        return base64.b64decode(r[7:])
    return r


def get_variables(refs, consolidated=True):
    """Get list of variable names from references.

    Finds the top-level prefixes in a reference set, corresponding to
    the directory listing of the root for zarr.

    Parameters
    ----------
    refs : dict
        kerchunk references keys
    consolidated : bool
        Whether or not to add consolidated metadata key to references. (default True)

    Returns
    -------
    fields : list of str
        List of variable names.
    """
    fields = []
    meta = {}
    for k in refs:
        if ".z" in k and consolidated:
            meta[k] = refs[k]
        if "/" in k:
            name, chunk = k.split("/")
            if name not in fields:
                fields.append(name)
            else:
                continue
        else:
            fields.append(k)
    if consolidated and ".zmetadata" not in fields:
        zarr.consolidate_metadata(meta)
        refs[".zmetadata"] = ujson.loads(meta[".zmetadata"])
        fields.append(".zmetadata")
    return fields


def _normalize_json(json_obj):
    """Normalize json representation as bytes

    Parameters
    ----------
    json_obj : str, bytes, dict, list
        JSON data for parquet file to be written.
    """
    if not isinstance(json_obj, str) and not isinstance(json_obj, bytes):
        json_obj = ujson.dumps(json_obj)
    if not isinstance(json_obj, bytes):
        json_obj = json_obj.encode()
    return json_obj


def _write_json(fname, json_obj, storage_options=None):
    """Write references into a parquet file.

    Parameters
    ----------
    fname : str
        Output filename.
    json_obj : str, bytes, dict, list
        JSON data for parquet file to be written.
    """
    storage_options = {} if storage_options is None else storage_options
    json_obj = _normalize_json(json_obj)
    with fsspec.open(fname, "wb", **storage_options) as f:
        f.write(json_obj)


def refs_to_dataframe(
    fo,
    url,
    target_protocol=None,
    target_options=None,
    storage_options=None,
    record_size=100_000,
    categorical_threshold=10,
):
    """Write references as a parquet files store.

    The directory structure should mimic a normal zarr store but instead of standard chunk
    keys, references are saved as parquet dataframes.

    Parameters
    ----------
    fo : str | dict
        Location of a JSON file containing references or a reference set already loaded
        into memory.
    url : str
        Location for the output, together with protocol. This must be a writable
        directory.
    target_protocol : str
        Used for loading the reference file, if it is a path. If None, protocol
        will be derived from the given path
    target_options : dict
        Extra FS options for loading the reference file ``fo``, if given as a path
    storage_options: dict | None
        Passed to fsspec when for writing the parquet.
    record_size : int
        Number of references to store in each reference file (default 10000). Bigger values
        mean fewer read requests but larger memory footprint.
    categorical_threshold : int
        Encode urls as pandas.Categorical to reduce memory footprint if the ratio
        of the number of unique urls to total number of refs for each variable
        is greater than or equal to this number. (default 10)
    """
    from fsspec.implementations.reference import LazyReferenceMapper

    if isinstance(fo, str):
        # JSON file
        dic = dict(**(target_options or {}), protocol=target_protocol)
        with fsspec.open(fo, "rb", **dic) as f:
            logger.info("Read reference from URL %s", fo)
            refs = ujson.load(f)
    else:
        # Mapping object
        refs = fo

    if "refs" in refs:
        refs = refs["refs"]

    fs, _ = fsspec.core.url_to_fs(url, **(storage_options or {}))
    out = LazyReferenceMapper.create(
        record_size=record_size,
        root=url,
        fs=fs,
        categorical_threshold=categorical_threshold,
    )

    for k in sorted(refs):
        out[k] = refs[k]
    out.flush()



import fsspec
import logging
import numcodecs
import numcodecs.abc
import numpy as np
import zarr

from fsspec.implementations.reference import LazyReferenceMapper


from kerchunk.utils import class_factory, dict_to_store, translate_refs_serializable
from kerchunk.codecs import AsciiTableCodec, VarArrCodec

try:
    from astropy.wcs import WCS
    from astropy.io import fits
except ModuleNotFoundError:  # pragma: no cover
    raise ImportError(
        "astropy is required for kerchunking FIST files. Please install with "
        "`pip/conda install astropy`."
    )

logger = logging.getLogger("fits-to-zarr")


BITPIX2DTYPE = {
    8: "uint8",
    16: ">i2",
    32: ">i4",
    64: ">i8",
    -32: ">f4",
    -64: ">f8",
}  # always bigendian


def process_file(
    url,
    storage_options=None,
    extension=None,
    inline_threshold=100,
    primary_attr_to_group=False,
    out=None,
):
    """
    Create JSON references for a single FITS file as a zarr group

    Parameters
    ----------
    url: str
        Where the file is
    storage_options: dict
        How to load that file (passed to fsspec)
    extension: list(int | str) | int | str or None
        Which extensions to include. Can be ordinal integer(s), the extension name (str) or if None,
        uses the first data extension
    inline_threshold: int
        (not yet implemented)
    primary_attr_to_group: bool
        Whether the output top-level group contains the attributes of the primary extension
        (which often contains no data, just a general description)
    out: dict-like or None
        This allows you to supply an fsspec.implementations.reference.LazyReferenceMapper
        to write out parquet as the references get filled, or some other dictionary-like class
        to customise how references get stored


    Returns
    -------
    dict of the references
    """
    from astropy.io import fits

    storage_options = storage_options or {}
    out = out or {}
    store = dict_to_store(out)
    g = zarr.open_group(store=store, zarr_format=2)

    with fsspec.open(url, mode="rb", **storage_options) as f:
        infile = fits.open(f, do_not_scale_image_data=True)
        if extension is None:
            found = False
            for i, hdu in enumerate(infile):
                if hdu.header.get("NAXIS", 0) > 0:
                    extension = [i]
                    found = True
                    break
            if not found:
                raise ValueError("No data extensions")
        elif isinstance(extension, (int, str)):
            extension = [extension]

        for ext in extension:
            hdu = infile[ext]
            hdu.header.__str__()  # causes fixing of invalid cards

            attrs = dict(hdu.header)
            kwargs = {"compressor": None}
            if hdu.is_image:
                # for images/cubes (i.e., ndarrays with simple type)
                nax = hdu.header["NAXIS"]
                shape = tuple(int(hdu.header[f"NAXIS{i}"]) for i in range(nax, 0, -1))
                dtype = BITPIX2DTYPE[hdu.header["BITPIX"]]
                length = np.dtype(dtype).itemsize
                for s in shape:
                    length *= s

                if "BSCALE" in hdu.header or "BZERO" in hdu.header:
                    kwargs["filters"] = [
                        numcodecs.FixedScaleOffset(
                            offset=float(hdu.header.get("BZERO", 0)),
                            scale=float(hdu.header.get("BSCALE", 1)),
                            astype=dtype,
                            dtype=float,
                        )
                    ]
                else:
                    kwargs["filters"] = None
            elif isinstance(hdu, fits.hdu.table.TableHDU):
                # ascii table
                spans = hdu.columns._spans
                outdtype = [
                    [name, hdu.columns[name].format.recformat]
                    for name in hdu.columns.names
                ]
                indtypes = [
                    [name, f"S{i + 1}"] for name, i in zip(hdu.columns.names, spans)
                ]
                nrows = int(hdu.header["NAXIS2"])
                shape = (nrows,)
                kwargs["filters"] = [AsciiTableCodec(indtypes, outdtype)]
                dtype = [tuple(d) for d in outdtype]
                length = (sum(spans) + len(spans)) * nrows
            elif isinstance(hdu, fits.hdu.table.BinTableHDU):
                # binary table
                dtype = hdu.columns.dtype.newbyteorder(">")  # always big endian
                nrows = int(hdu.header["NAXIS2"])
                shape = (nrows,)
                # if hdu.fileinfo()["datSpan"] > length
                if any(_.format.startswith(("P", "Q")) for _ in hdu.columns):
                    # contains var fields
                    length = hdu.fileinfo()["datSpan"]
                    dt2 = [
                        (
                            (name, "O")
                            if hdu.columns[name].format.startswith(("P", "Q"))
                            else (name, str(dtype[name].base))
                            + ((dtype[name].shape,) if dtype[name].shape else ())
                        )
                        for name in dtype.names
                    ]
                    types = {
                        name: hdu.columns[name].format[1]
                        for name in dtype.names
                        if hdu.columns[name].format.startswith(("P", "Q"))
                    }
                    kwargs["compressor"] = VarArrCodec(
                        str(dtype), str(dt2), nrows, types
                    )
                    kwargs["fill_value"] = None
                    dtype = dt2
                else:
                    length = dtype.itemsize * nrows
                    kwargs["filters"] = None
            else:
                logger.warning(f"Skipping unknown extension type: {hdu}")
                continue
            # one chunk for whole thing.
            # TODO: we could sub-chunk on biggest dimension
            name = hdu.name or str(ext)
            arr = g.create_array(
                name=name, dtype=dtype, shape=shape, chunks=shape, **kwargs
            )
            arr.attrs.update(
                {
                    k: str(v) if not isinstance(v, (int, float, str)) else v
                    for k, v in attrs.items()
                    if k != "COMMENT"
                }
            )
            arr.attrs["_ARRAY_DIMENSIONS"] = ["z", "y", "x"][-len(shape) :]
            loc = hdu.fileinfo()["datLoc"]
            parts = ".".join(["0"] * len(shape))
            out[f"{name}/{parts}"] = [url, loc, length]
        if primary_attr_to_group:
            # copy attributes of primary extension to top-level group
            hdu = infile[0]
            hdu.header.__str__()
            g.attrs.update(
                {
                    k: str(v) if not isinstance(v, (int, float, str)) else v
                    for k, v in dict(hdu.header).items()
                    if k != "COMMENT"
                }
            )
    if isinstance(out, LazyReferenceMapper):
        out.flush()
    out = translate_refs_serializable(out)
    return out


FitsToZarr = class_factory(process_file)


def add_wcs_coords(hdu, zarr_group=None, dataset=None, dtype="float32"):
    """Using FITS WCS, create materialised coordinate arrays

    This may triple the data footprint of the data, as the coordinates can easily
    be as big as the data itsel.

    Must provide zarr_group or dataset

    Parameters
    ----------
    hdu: astropy.io.fits.HDU or dict
        Input with WCS header information. If a dict, it is {key: attribute} of the data.
    zarr_group: zarr.Group
        To write the new arrays into
    dataset: xr.Dataset
        To create new coordinate arrays in; this is not necessarily written anywhere
    dtype: str
        Output numpy dtype

    Returns
    -------
    If dataset is given, returns the modified dataset.
    """
    if zarr_group is None and dataset is None:
        raise ValueError("please provide a zarr group or xarray dataset")

    if isinstance(hdu, dict):
        # assume dict-like
        head = fits.Header()
        hdu2 = hdu.copy()
        hdu2.pop("COMMENT", None)  # comment fields can be non-standard
        head.update(hdu2)
        hdu = fits.PrimaryHDU(header=head)
    elif not isinstance(hdu, fits.hdu.base._BaseHDU):
        raise TypeError("`hdu` must be a FITS HDU or dict")
    nax = hdu.header["NAXIS"]
    shape = tuple(int(hdu.header[f"NAXIS{i}"]) for i in range(nax, 0, -1))

    wcs = WCS(hdu)
    coords = [
        coo.ravel() for coo in np.meshgrid(*(np.arange(sh) for sh in shape))
    ]  # ?[::-1]
    world_coords = wcs.pixel_to_world(*coords)
    for i, (name, world_coord) in enumerate(zip(wcs.axis_type_names, world_coords)):
        dims = ["z", "y", "x"][3 - len(shape) :]
        attrs = {
            "unit": world_coord.unit.name,
            "type": hdu.header[f"CTYPE{i + 1}"],
            "_ARRAY_DIMENSIONS": dims,
        }
        if zarr_group is not None:
            arr = zarr_group.empty(
                name, shape=shape, chunks=shape, dtype=dtype, exists_ok=True
            )
            arr.attrs.update(attrs)
            arr[:] = world_coord.value.reshape(shape)
        if dataset is not None:
            import xarray as xr

            coo = xr.Coordinate(
                data=world_coord.value.reshape(shape), dims=dims, attrs=attrs
            )
            dataset = dataset.assign_coordinates(name=coo)
    if dataset is not None:
        return dataset



import xarray as xr
import numpy as np
from kerchunk import netCDF3

arr = np.random.rand(1, 10, 10)
data = xr.DataArray(
    data=arr.squeeze(),
    dims=["x", "y"],
    name="data",
)
bdata = xr.Dataset({"data": data}, attrs={"attr0": 3}).to_netcdf(
    format="NETCDF3_CLASSIC"
)


# Add parameterize test for storage formats (.json, .parquet)
def test_reference_netcdf(m):
    m.pipe("data.nc3", bdata)
    h = netCDF3.netcdf_recording_file("memory://data.nc3")
    out = h.translate()
    ds = xr.open_dataset(
        out, engine="kerchunk", storage_options={"remote_protocol": "memory"}
    )
    assert (ds.data == data).all()



import pytest
import re

import fsspec
import xarray as xr

from kerchunk.combine import auto_dask
from kerchunk.utils import refs_as_store
from kerchunk.zarr import ZarrToZarr

dask = pytest.importorskip("dask")


@pytest.mark.parametrize("n_batches", [1, 2, 3])
def test_simplest(m, n_batches):
    for i in range(4):
        m.pipe(
            {
                f"data{i}/.zgroup": b'{"zarr_format":2}',
                f"data{i}/data/.zarray": b'{"chunks":[3],"compressor": null,"dtype": "|i1",'
                b'"fill_value": 0,"filters": null,"order": "C",'
                b'"shape": [3],"zarr_format": 2}',
                f"data{i}/data/0": f"{i}{i}{i}".encode(),
            }
        )
    out = auto_dask(
        [f"memory:///data{i}" for i in range(4)],
        single_driver=ZarrToZarr,
        single_kwargs={"inline": 0},
        n_batches=n_batches,
        mzz_kwargs={
            "coo_map": {"count": re.compile(r".*(\d)")},
            "inline_threshold": 0,
            "coo_dtypes": {"count": "i4"},
        },
    )
    ds = xr.open_dataset(
        refs_as_store(out), engine="zarr", backend_kwargs={"consolidated": False}
    )
    assert ds["count"].values.tolist() == [0, 1, 2, 3]
    assert ds.data.shape == (4, 3)
    assert (ds.data.values.T == [48, 49, 50, 51]).all()



import fsspec
from fsspec.implementations.reference import LazyReferenceMapper

import kerchunk.utils
import ujson


def single_zarr(
    uri_or_store,
    storage_options=None,
    inline_threshold=100,
    inline=None,
    out=None,
):
    """kerchunk-style view on zarr mapper

    This is a similar process to zarr's consolidate_metadata, but does not
    need to be held in the original file tree. You do not need zarr itself
    to do this.

    This is useful for testing, so that we can pass hand-made zarrs to combine.

    Parameters
    ----------
    uri_or_store: str or dict-like
    storage_options: dict or None
        given to fsspec
    out: dict-like or None
        This allows you to supply an fsspec.implementations.reference.LazyReferenceMapper
        to write out parquet as the references get filled, or some other dictionary-like class
        to customise how references get stored

    Returns
    -------
    reference dict like
    """
    if isinstance(uri_or_store, str):
        mapper = fsspec.get_mapper(uri_or_store, **(storage_options or {}))
        prot = mapper.fs.protocol
        protocol = prot[0] if isinstance(prot, tuple) else prot
    else:
        mapper = uri_or_store
        if isinstance(mapper, fsspec.FSMap) and storage_options is None:
            storage_options = mapper.fs.storage_options
            prot = mapper.fs.protocol
            protocol = prot[0] if isinstance(prot, tuple) else prot
        else:
            protocol = None

    try:
        check = ujson.loads(mapper[".zgroup"])
        assert check["zarr_format"] == 2
    except (KeyError, ValueError, TypeError) as e:
        raise ValueError("Failed to load dataset as V2 zarr") from e
    refs = out or {}
    for k in mapper:
        if k.startswith("."):
            refs[k] = mapper[k]
        else:
            refs[k] = [fsspec.utils._unstrip_protocol(mapper._key_to_str(k), mapper.fs)]
    from kerchunk.utils import do_inline

    inline_threshold = inline if inline is not None else inline_threshold
    refs = do_inline(
        refs,
        inline_threshold,
        remote_options=storage_options,
        remote_protocol=protocol,
    )
    if isinstance(refs, LazyReferenceMapper):
        refs.flush()
    refs = kerchunk.utils.consolidate(refs)
    return refs


ZarrToZarr = kerchunk.utils.class_factory(single_zarr)



import io

import fsspec
import json
import kerchunk.utils
import kerchunk.zarr
import numpy as np
import pytest
import zarr

from fsspec.implementations.asyn_wrapper import AsyncFileSystemWrapper


def test_rename():
    old = {"version": 1, "refs": {"v0": ["oldpath", 0, 0], "bin": "data"}}
    new = kerchunk.utils.rename_target(old, {"oldpath": "newpath"})
    assert new == {"version": 1, "refs": {"v0": ["newpath", 0, 0], "bin": "data"}}


def test_rename_files(m):
    m.pipe(
        "ref.json", b'{"version": 1, "refs": {"v0": ["oldpath", 0, 0], "bin": "data"}}'
    )
    kerchunk.utils.rename_target_files("memory://ref.json", {"oldpath": "newpath"})
    out = m.cat("ref.json")
    assert out == b'{"version":1,"refs":{"v0":["newpath",0,0],"bin":"data"}}'
    kerchunk.utils.rename_target_files(
        "memory://ref.json", {"newpath": "newerpath"}, url_out="memory://ref2.json"
    )
    out = m.cat("ref.json")
    assert out == b'{"version":1,"refs":{"v0":["newpath",0,0],"bin":"data"}}'
    out = m.cat("ref2.json")
    assert out == b'{"version":1,"refs":{"v0":["newerpath",0,0],"bin":"data"}}'


def test_inline(m):
    m.pipe("data", b"stuff")
    refs = {
        "key0": b"00",
        "key1": ["memory://data"],
        "key2": ["memory://data", 1, 1],
        "key1": ["memory://data", 2, 4],
    }
    out = kerchunk.utils.do_inline(refs, 2)
    expected = {
        "key0": b"00",
        "key1": ["memory://data"],
        "key2": b"t",
        "key1": ["memory://data", 2, 4],
    }
    assert out == expected


def test_inline_array():
    refs = {
        ".zgroup": b'{"zarr_format": 2}',
        "data/.zarray": """
    {
    "chunks": [
        1
    ],
    "compressor": null,
    "dtype": "<i4",
    "fill_value": 0,
    "filters": null,
    "order": "C",
    "shape": [
        2
    ],
    "zarr_format": 2
}
""",
        "data/0": b"\x01\x00\x00\x00",
        "data/1": b"\x02\x00\x00\x00",
        "data/.zattrs": '{"foo": "bar"}',
    }
    out1 = kerchunk.utils.inline_array(refs, threshold=1)  # does nothing
    assert out1 == refs
    out2 = kerchunk.utils.inline_array(refs, threshold=1, names=["data"])  # explicit
    assert json.loads(out2["data/.zattrs"]) == json.loads(refs["data/.zattrs"])

    localfs = fsspec.filesystem("file")
    store = kerchunk.utils.refs_as_store(out2, fs=localfs)
    g = zarr.open(store, mode="r", zarr_format=2)
    assert g["data"][:].tolist() == [1, 2]  # What is g.data???

    out3 = kerchunk.utils.inline_array(refs, threshold=1000)  # inlines because of size
    store = kerchunk.utils.refs_as_store(out3, localfs)
    g = zarr.open(store, mode="r", zarr_format=2)
    assert g["data"][:].tolist() == [1, 2]  # What is g.data???


def test_json():
    data = {"a": "a", "b": b"b", "c": [None, None, None], "d": '{"key": 0}'}
    out = kerchunk.utils._encode_for_JSON(data)
    expected = {"a": "a", "b": "b", "c": [None, None, None], "d": '{"key":0}'}
    assert out == expected


@pytest.mark.parametrize("chunks", [[10, 10], [5, 10]])
def test_subchunk_exact(m, chunks):
    g = zarr.open_group("memory://test.zarr", mode="w", zarr_format=2)
    data = np.arange(100).reshape(10, 10)
    arr = g.create_array(
        "data", dtype=data.dtype, shape=data.shape, chunks=chunks, compressor=None
    )
    arr[:] = data
    ref = kerchunk.zarr.single_zarr("memory://test.zarr")["refs"]

    extra = [] if chunks[0] == 10 else ["data/1.0"]
    ref2 = list(_ for _ in ref if not _.endswith("zattrs"))  # ignore empty attrs
    assert ref2 == [".zgroup", "data/.zarray", "data/0.0"] + extra

    out = kerchunk.utils.subchunk(ref, "data", 5)
    nchunk = 10 // chunks[0] * 5
    out2 = list(_ for _ in out if not _.endswith("zattrs"))
    assert out2 == [".zgroup", "data/.zarray"] + [f"data/{_}.0" for _ in range(nchunk)]

    store = kerchunk.utils.refs_as_store(out, remote_protocol="memory")
    g2 = zarr.open_group(store, mode="r", zarr_format=2)

    # g2 = zarr.open_group(
    #     "reference://", storage_options={"fo": out, "remote_protocol": "memory"}, zarr_format=2
    # )
    assert (g2["data"][:] == data).all()


@pytest.mark.parametrize("archive", ["zip", "tar"])
def test_archive(m, archive):
    import zipfile
    import tarfile

    data = b"piece of data"
    with fsspec.open("memory://archive", "wb") as f:
        if archive == "zip":
            arc = zipfile.ZipFile(file=f, mode="w")
            arc.writestr("data1", data)
            arc.close()
        else:
            arc = tarfile.TarFile(fileobj=f, mode="w")
            ti = tarfile.TarInfo("data1")
            ti.size = len(data)
            arc.addfile(ti, io.BytesIO(data))
            arc.close()
    refs = {
        "a": b"stuff",
        "b": [f"{archive}://data1::memory://archive"],
        "c": [f"{archive}://data1::memory://archive", 5, 2],
    }

    refs2 = kerchunk.utils.dereference_archives(refs)

    fs = fsspec.filesystem("reference", fo=refs2)
    assert fs.cat("a") == b"stuff"
    assert fs.cat("b") == data
    assert fs.cat("c") == data[5:7]


def test_deflate_zip_archive(m):
    import zipfile
    from kerchunk.codecs import DeflateCodec

    dec = DeflateCodec()

    data = b"piece of data"
    with fsspec.open("memory://archive", "wb") as f:
        arc = zipfile.ZipFile(file=f, mode="w", compression=zipfile.ZIP_DEFLATED)
        arc.writestr("data1", data)
        arc.close()
    refs = {
        "b": [f"zip://data1::memory://archive"],
    }

    with pytest.warns(UserWarning):
        refs2 = kerchunk.utils.dereference_archives(refs)

    fs = fsspec.filesystem("reference", fo=refs2)
    assert dec.decode(fs.cat("b")) == data


def test_encode_fill_value():
    assert kerchunk.utils.encode_fill_value(np.array([9999]), np.dtype("int")) == 9999
    assert kerchunk.utils.encode_fill_value(np.array(9999), np.dtype("int")) == 9999
    assert kerchunk.utils.encode_fill_value([9999], np.dtype("int")) == 9999
    assert kerchunk.utils.encode_fill_value(9999, np.dtype("int")) == 9999



import os.path

import eccodes
import fsspec
import numpy as np
import pandas as pd
import pytest
import xarray as xr
import zarr
import ujson
from fsspec.implementations.asyn_wrapper import AsyncFileSystemWrapper
from kerchunk.grib2 import (
    scan_grib,
    _split_file,
    GribToZarr,
    grib_tree,
    correct_hrrr_subhf_step,
    parse_grib_idx,
)
from kerchunk._grib_idx import (
    extract_dataset_chunk_index,
    extract_datatree_chunk_index,
)
from kerchunk.utils import fs_as_store, refs_as_store

eccodes_ver = tuple(int(i) for i in eccodes.__version__.split("."))
cfgrib = pytest.importorskip("cfgrib")
here = os.path.dirname(__file__)


def test_one():
    # from https://dd.weather.gc.ca/model_gem_regional/10km/grib2/00/000
    fn = os.path.join(here, "CMC_reg_DEPR_ISBL_10_ps10km_2022072000_P000.grib2")
    out = scan_grib(fn)

    fs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    store = refs_as_store(out[0], fs=fs)
    ds = xr.open_zarr(store, zarr_format=2, consolidated=False)

    assert ds.attrs["GRIB_centre"] == "cwao"
    ds2 = xr.open_dataset(fn, engine="cfgrib", backend_kwargs={"indexpath": ""})

    for var in ["latitude", "longitude", "unknown", "isobaricInhPa", "time"]:
        d1 = ds[var].values
        d2 = ds2[var].values
        assert (np.isnan(d1) == np.isnan(d2)).all()
        assert (d1[~np.isnan(d1)] == d2[~np.isnan(d2)]).all()


def _fetch_first(url):
    fs = fsspec.filesystem("s3", anon=True)
    with fs.open(url, "rb") as f:
        for _, _, data in _split_file(f, skip=1):
            return data


@pytest.mark.parametrize(
    "url",
    [
        pytest.param(
            "s3://noaa-hrrr-bdp-pds/hrrr.20140730/conus/hrrr.t23z.wrfsubhf08.grib2",
            marks=pytest.mark.skipif(eccodes_ver >= (2, 34), reason="eccodes too new"),
        ),
        "s3://noaa-gefs-pds/gefs.20221011/00/atmos/pgrb2ap5/gep01.t00z.pgrb2a.0p50.f570",
        "s3://noaa-gefs-retrospective/GEFSv12/reforecast/2000/2000010100/c00/Days:10-16/acpcp_sfc_2000010100_c00.grib2",
    ],
)
def test_archives(tmpdir, url):
    grib = GribToZarr(url, storage_options={"anon": True}, skip=1)
    out = grib.translate()[0]

    store = refs_as_store(out, remote_options={"anon": True})

    ours = xr.open_zarr(store, zarr_format=2, consolidated=False)

    data = _fetch_first(url)
    fn = os.path.join(tmpdir, "grib.grib2")
    with open(fn, "wb") as f:
        f.write(data)

    theirs = cfgrib.open_dataset(fn)
    if "hrrr" in url:
        # for some reason, cfgrib reads `step` as 7.25 hours
        # while grib_ls and kerchunk reads `step` as 425 hours.
        ours = ours.drop_vars("step")
        theirs = theirs.drop_vars("step")

    xr.testing.assert_allclose(ours, theirs)


def test_subhourly():
    # two messages extracted from a hrrr output including one with an eccodes
    # non-compliant endstep type which raises WrongStepUnitError
    fpath = os.path.join(here, "hrrr.wrfsubhf.sample.grib2")
    result = scan_grib(fpath)
    assert len(result) == 2, "Expected two grib messages"


def test_tiny_grib():
    fpath = os.path.join(here, "tinygrib.grb2")
    result = scan_grib(fpath)
    assert len(result) == 1, "Expected one grib message"


def test_grib_tree():
    """
    End-to-end test from grib file to zarr hierarchy
    """
    fpath = os.path.join(here, "hrrr.wrfsubhf.sample.grib2")
    scanned_msg_groups = scan_grib(fpath)
    corrected_msg_groups = [correct_hrrr_subhf_step(msg) for msg in scanned_msg_groups]
    result = grib_tree(corrected_msg_groups)
    fs = fsspec.filesystem("reference", fo=result)
    store = fs_as_store(fs)
    zg = zarr.open_group(store, mode="r", zarr_format=2)
    assert isinstance(zg["refc/instant/atmosphere/refc"], zarr.Array)
    assert isinstance(zg["vbdsf/avg/surface/vbdsf"], zarr.Array)
    assert set(zg["vbdsf/avg/surface"].attrs["coordinates"].split()) == set(
        "surface latitude longitude step time valid_time".split()
    )
    assert set(zg["refc/instant/atmosphere"].attrs["coordinates"].split()) == set(
        "atmosphere latitude longitude step time valid_time".split()
    )
    # Assert that the fill value is set correctly
    assert np.isnan(zg["refc/instant/atmosphere/step"].fill_value)


# The following two tests use json fixture data generated from calling scan grib
#   scan_grib("testdata/hrrr.t01z.wrfsubhf00.grib2")
#   scan_grib("testdata/hrrr.t01z.wrfsubhf01.grib2")
# and filtering the results message groups for keys starting with "dswrf" or "u"
# The original files are:
# gs://high-resolution-rapid-refresh/hrrr.20210928/conus/hrrr.t01z.wrfsubhf00.grib2"
# gs://high-resolution-rapid-refresh/hrrr.20210928/conus/hrrr.t01z.wrfsubhf01.grib2"


def test_correct_hrrr_subhf_group_step():
    fpath = os.path.join(here, "hrrr.wrfsubhf.subset.json")
    with open(fpath, "rb") as fobj:
        scanned_msgs = ujson.load(fobj)

    original_zg = [
        zarr.open_group(
            fs_as_store(fsspec.filesystem("reference", fo=val)), mode="r", zarr_format=2
        )
        for val in scanned_msgs
    ]

    corrected_msgs = [correct_hrrr_subhf_step(msg) for msg in scanned_msgs]

    corrected_zg = [
        zarr.open_group(
            fs_as_store(fsspec.filesystem("reference", fo=val)), mode="r", zarr_format=2
        )
        for val in corrected_msgs
    ]

    # The groups that were missing a step variable got fixed
    assert all(["step" in zg.array_keys() for zg in corrected_zg])
    assert not all(["step" in zg.array_keys() for zg in original_zg])

    # The step values are corrected to floating point hour
    assert all([zg["step"][()] <= 1.0 for zg in corrected_zg])
    # The original seems to have values in minutes for some step variables!
    assert not all(
        [zg["step"][()] <= 1.0 for zg in original_zg if "step" in zg.array_keys()]
    )


def test_hrrr_subhf_corrected_grib_tree():
    fpath = os.path.join(here, "hrrr.wrfsubhf.subset.json")
    with open(fpath, "rb") as fobj:
        scanned_msgs = ujson.load(fobj)

    corrected_msgs = [correct_hrrr_subhf_step(msg) for msg in scanned_msgs]
    merged = grib_tree(corrected_msgs)
    z_fs = fsspec.filesystem("reference", fo=merged, asynchronous=True)
    zstore = fs_as_store(z_fs)
    zg = zarr.open_group(zstore, mode="r", zarr_format=2)
    # Check the values and shape of the time coordinates
    assert zg["u/instant/heightAboveGround/step"][:].tolist() == [
        0.0,
        0.25,
        0.5,
        0.75,
        1.0,
    ]
    assert zg["u/instant/heightAboveGround/step"].shape == (5,)
    assert zg["u/instant/heightAboveGround/valid_time"][:].tolist() == [
        [1695862800, 1695863700, 1695864600, 1695865500, 1695866400]
    ]
    assert zg["u/instant/heightAboveGround/valid_time"].shape == (1, 5)
    assert zg["u/instant/heightAboveGround/time"][:].tolist() == [1695862800]
    assert zg["u/instant/heightAboveGround/time"].shape == (1,)
    assert zg["dswrf/avg/surface/step"][:].tolist() == [0.0, 0.25, 0.5, 0.75, 1.0]
    assert zg["dswrf/avg/surface/step"].shape == (5,)
    assert zg["dswrf/avg/surface/valid_time"][:].tolist() == [
        [1695862800, 1695863700, 1695864600, 1695865500, 1695866400]
    ]
    assert zg["dswrf/avg/surface/valid_time"].shape == (1, 5)
    assert zg["dswrf/avg/surface/time"][:].tolist() == [1695862800]
    assert zg["dswrf/avg/surface/time"].shape == (1,)


# The following two test use json fixture data generated from calling scan grib
#   scan_grib("testdata/hrrr.t01z.wrfsfcf00.grib2")
#   scan_grib("testdata/hrrr.t01z.wrfsfcf01.grib2")
# and filtering the results for keys starting with "dswrf" or "u"
# The original files are:
# gs://high-resolution-rapid-refresh/hrrr.20210928/conus/hrrr.t01z.wrfsfcf00.grib2"
# gs://high-resolution-rapid-refresh/hrrr.20210928/conus/hrrr.t01z.wrfsfcf01.grib2"
def test_hrrr_sfcf_grib_tree():
    fpath = os.path.join(here, "hrrr.wrfsfcf.subset.json")
    with open(fpath, "rb") as fobj:
        scanned_msgs = ujson.load(fobj)
    merged = grib_tree(scanned_msgs)
    store = fs_as_store(fsspec.filesystem("reference", fo=merged))
    zg = zarr.open_group(store, mode="r", zarr_format=2)
    # Check the heightAboveGround level shape of the time coordinates
    assert zg["u/instant/heightAboveGround/heightAboveGround"][()] == 80.0
    assert zg["u/instant/heightAboveGround/heightAboveGround"].shape == ()
    assert zg["u/instant/heightAboveGround/step"][:].tolist() == [0.0, 1.0]
    assert zg["u/instant/heightAboveGround/step"].shape == (2,)
    assert zg["u/instant/heightAboveGround/valid_time"][:].tolist() == [
        [1695862800, 1695866400]
    ]
    assert zg["u/instant/heightAboveGround/valid_time"].shape == (1, 2)
    assert zg["u/instant/heightAboveGround/time"][:].tolist() == [1695862800]
    assert zg["u/instant/heightAboveGround/time"].shape == (1,)

    # Check the isobaricInhPa level shape and time coordinates
    assert zg["u/instant/isobaricInhPa/isobaricInhPa"][:].tolist() == [
        250.0,
        300.0,
        500.0,
        700.0,
        850.0,
        925.0,
        1000.0,
    ]
    assert zg["u/instant/isobaricInhPa/isobaricInhPa"].shape == (7,)
    assert zg["u/instant/isobaricInhPa/step"][:].tolist() == [0.0, 1.0]
    assert zg["u/instant/isobaricInhPa/step"].shape == (2,)

    # Valid time values get exploded by isobaricInhPa aggregation
    # Is this a feature or a bug?
    expected_valid_times = [
        [
            [1695862800 for _ in range(7)],
            [1695866400 for _ in range(7)],
        ]
    ]
    assert zg["u/instant/isobaricInhPa/valid_time"][:].tolist() == expected_valid_times
    assert zg["u/instant/isobaricInhPa/valid_time"].shape == (1, 2, 7)

    assert zg["u/instant/isobaricInhPa/time"][:].tolist() == [1695862800]
    assert zg["u/instant/isobaricInhPa/time"].shape == (1,)


# def test_hrrr_sfcf_grib_datatree():
#     fpath = os.path.join(here, "hrrr.wrfsfcf.subset.json")
#     with open(fpath, "rb") as fobj:
#         scanned_msgs = ujson.load(fobj)
#     merged = grib_tree(scanned_msgs)
#     dt = datatree.open_datatree(
#         fsspec.filesystem("reference", fo=merged).get_mapper(""),
#         engine="zarr",
#         consolidated=False,
#     )
#     # Assert a few things... but if it loads we are mostly done.
#     np.testing.assert_array_equal(
#         dt.u.instant.heightAboveGround.step.values[:],
#         np.array([0, 3600 * 10**9], dtype="timedelta64[ns]"),
#     )
#     assert dt.u.attrs == dict(name="U component of wind")


def test_parse_grib_idx_invalid_url():
    with pytest.raises(ValueError):
        # a random protocol is used
        parse_grib_idx(
            "ds://global-forecast-system/gfs.20230928/00/atmos/gfs.t00z.pgrb2.0p25.f001"
        )


def test_parse_grib_idx_no_file():
    with pytest.raises(PermissionError):
        # the url is spelled wrong
        parse_grib_idx(
            "s3://noaahrrr-bdp-pds/hrrr.20220804/conus/hrrr.t01z.wrfsfcf01.grib2",
            storage_options={"anon": True},
        )


def test_parse_grib_idx_duplicate_attrs():
    fn = os.path.join(here, "gfs.t00z.pgrb2.0p25.f006.test-limit-100")
    with pytest.raises(
        ValueError, match=f"Attribute mapping for grib file {fn} is not unique"
    ):
        parse_grib_idx(fn, validate=True)


@pytest.mark.parametrize(
    "idx_url, storage_options",
    [
        (
            "gs://global-forecast-system/gfs.20230928/00/atmos/gfs.t00z.pgrb2.0p25.f001",
            dict(),
        ),
        (
            "s3://noaa-gefs-pds/gefs.20170101/06/gec00.t06z.pgrb2af024",
            dict(anon=True),
        ),
    ],
)
# the test will fail in case the network goes down or the file is moved
def test_parse_grib_idx_content(idx_url, storage_options):
    import re

    # s3 variable assuming you don't have gcsfs locally
    variable = "gh/0.0"

    if re.match(r"gs://|gcs://", idx_url):
        pytest.importorskip("gcsfs", reason="gcsfs is not installed on the system")
        variable = "prmsl/0.0"

    idx_df = parse_grib_idx(idx_url, storage_options=storage_options)
    assert isinstance(idx_df, pd.DataFrame)

    # comparing the url, length, offset values of the first message
    message_no = 0
    output = scan_grib(idx_url, skip=15, storage_options=storage_options)
    assert idx_df.iloc[message_no]["grib_uri"] == output[message_no]["templates"]["u"]

    assert idx_df.iloc[message_no]["offset"] == output[message_no]["refs"][variable][1]
    assert idx_df.iloc[message_no]["length"] == output[message_no]["refs"][variable][2]


# @pytest.fixture
# def zarr_tree_and_datatree_instance():
#     fn = os.path.join(here, "gfs.t00z.pgrb2.0p25.f006.test-limit-100")
#     tree_store = tree_store = grib_tree(scan_grib(fn))
#     dt_instance = datatree.open_datatree(
#         fsspec.filesystem("reference", fo=tree_store).get_mapper(""),
#         engine="zarr",
#         consolidated=False,
#     )

#     return tree_store, dt_instance, fn


@pytest.mark.skip(reason="datatree support should be updated to use xarray.Datatree")
def test_extract_dataset_chunk_index(zarr_tree_and_datatree_instance):
    tree_store, dt_instance, fn = zarr_tree_and_datatree_instance

    # extracting the metadata from a single datatree node with key "t/instant/isobaricInhPa" and comparing it.
    idx_metadata = extract_dataset_chunk_index(
        dt_instance["t/instant/isobaricInhPa"], tree_store["refs"]
    )

    # comparing the first metadata generated from the datatree zarr store
    assert idx_metadata[0]["uri"] == fn
    assert idx_metadata[0]["offset"] == 5130730
    assert idx_metadata[0]["length"] == 844711

    # comparing the first metadata with the datatree instance
    assert (
        idx_metadata[0]["valid_time"]
        == dt_instance["t/instant/isobaricInhPa"]
        .coords["valid_time"][0][0][0]
        .to_numpy()
    )
    assert (
        idx_metadata[0]["isobaricInhPa"]
        == dt_instance["t/instant/isobaricInhPa"].coords["isobaricInhPa"][0].to_numpy()
    )
    assert (
        idx_metadata[0]["step"]
        == dt_instance["t/instant/isobaricInhPa"].coords["step"][0].to_numpy()
    )


@pytest.mark.skip(reason="datatree support should be updated to use xarray.Datatree")
def test_extract_datatree_chunk_index(zarr_tree_and_datatree_instance):
    tree_store, dt_instance, fn = zarr_tree_and_datatree_instance

    # dataframe generated from a single grib file
    idx_df = extract_datatree_chunk_index(dt_instance, tree_store)

    assert isinstance(idx_df, pd.DataFrame)

    # comparing the dataframe's columns
    assert len(idx_df.columns) == 15
    assert (
        idx_df.columns.to_numpy()
        == [
            "varname",
            "typeOfLevel",
            "stepType",
            "name",
            "isobaricInhPa",
            "step",
            "time",
            "valid_time",
            "uri",
            "offset",
            "length",
            "inline_value",
            "surface",
            "heightAboveGround",
            "meanSea",
        ]
    ).all()

    # comparing the actual values from random places in the dataframe
    assert (idx_df["varname"][[1, 4, 62, 75]] == ["absv", "acpcp", "t", "v"]).all()
    assert (
        idx_df["typeOfLevel"][[42, 46, 49, 50]]
        == ["heightAboveGround", "isobaricInhPa", "surface", "heightAboveGround"]
    ).all()
    assert (idx_df["stepType"][[26, 33, 83]] == ["instant", "avg", "accum"]).all()
    assert (idx_df["isobaricInhPa"][[17, 72, 85]] == [1000.0, 975.0, 975.0]).all()
    assert pd.isna(idx_df["isobaricInhPa"][56])
    assert (idx_df["step"] == np.timedelta64(21600000000000, "ns")).all()
    assert (
        idx_df["valid_time"] == np.datetime64("2023-09-28T06:00:00.000000000")
    ).all()
    assert (idx_df["uri"] == fn).all()
    assert (
        idx_df["offset"][[23, 45, 67, 80]] == [2243950, 14113893, 28609546, 8510703]
    ).all()
    assert (
        idx_df["length"][[12, 34, 78, 80]] == [722563, 576527, 947435, 1176904]
    ).all()


@pytest.mark.skip(reason="datatree support should be updated to use xarray.Datatree")
def test_extract_methods_grib_parameter(zarr_tree_and_datatree_instance):
    tree_store, dt_instance, _ = zarr_tree_and_datatree_instance

    # testing grib parameter with "extract_dataset_chunk_index" function
    grib_metadata = extract_dataset_chunk_index(
        dt_instance["t/instant/isobaricInhPa"], tree_store["refs"], True
    )

    # testing grib parameter with "extract_datatree_chunk_index" function
    metadata_df = extract_datatree_chunk_index(dt_instance, tree_store, True)

    # checking if level in the generated dataframe
    assert "level" in metadata_df.columns

    # checking if level in each series data
    assert all(list(map(lambda data: "level" in data.keys(), grib_metadata)))



import numpy
import h5py
import hdf5plugin

compressors = dict(
    zstd=hdf5plugin.Zstd(),
    bitshuffle=hdf5plugin.Bitshuffle(nelems=0, cname="lz4"),
    lz4=hdf5plugin.LZ4(nbytes=0),
    blosc_lz4_bitshuffle=hdf5plugin.Blosc(
        cname="blosclz", clevel=9, shuffle=hdf5plugin.Blosc.BITSHUFFLE
    ),
)

for c in compressors:
    f = h5py.File(f"hdf5_compression_{c}.h5", "w")
    # Explicit set int64, to keep compatible with Numpy 1.x in Windows
    f.create_dataset(
        "data", data=numpy.arange(100, dtype=numpy.int64), **compressors[c]
    )
    f.close()



import pandas as pd
import pytest

fastparquet = pytest.importorskip("fastparquet")
import fsspec
import ujson

from kerchunk.df import refs_to_dataframe


@pytest.mark.parametrize("url", [True, False])
def test_1(m, url):
    refs = {
        "a/0": ["memory://url1.file"],
        "a/1": ["memory://url1.file", 10, 100],
        "a/2": ["memory://url2.file"],
        "a/3": ["memory://url3.file"],
        "a/4": ["memory://url4.file"],
        "a/5": ["memory://url5.file"],
        "a/6": b"data",
        "a/.zarray": b"""{"shape": [7], "chunks":[1], "filters": [], "compressor": null}""",
        ".zgroup": b'{"zarr_format": 2}',
    }
    u = "memory://myrefs.json"
    if url:
        with fsspec.open(u, "wt") as f:
            ujson.dump(
                {
                    k: (v.decode() if isinstance(v, bytes) else v)
                    for k, v in refs.items()
                },
                f,
            )
        refs = u

    refs_to_dataframe(refs, "memory://outdir", record_size=4)
    with fsspec.open("memory:///outdir/.zmetadata") as f:
        meta = ujson.load(f)
        assert list(meta) == ["metadata", "record_size"]
        assert meta["record_size"] == 4

    df0 = pd.read_parquet("memory:///outdir/a/refs.0.parq")

    # no raw
    assert df0.to_dict() == {
        "offset": {0: 0, 1: 10, 2: 0, 3: 0},
        "path": {
            0: "memory://url1.file",
            1: "memory://url1.file",
            2: "memory://url2.file",
            3: "memory://url3.file",
        },
        "size": {0: 0, 1: 100, 2: 0, 3: 0},
        "raw": {0: None, 1: None, 2: None, 3: None},
    }

    # with raw column
    df1 = pd.read_parquet("memory:///outdir/a/refs.1.parq")
    assert df1.to_dict() == {
        "offset": {0: 0, 1: 0, 2: 0, 3: 0},
        "path": {0: "memory://url4.file", 1: "memory://url5.file", 2: None, 3: None},
        "raw": {0: None, 1: None, 2: b"data", 3: None},
        "size": {0: 0, 1: 0, 2: 0, 3: 0},
    }



from xarray.backends import BackendEntrypoint
import xarray as xr
import os

from kerchunk.utils import refs_as_store


class KerchunkBackend(BackendEntrypoint):
    def open_dataset(
        self, filename_or_obj, *, storage_options=None, open_dataset_options=None, **kw
    ):
        open_dataset_options = (open_dataset_options or {}) | kw
        ref_ds = open_reference_dataset(
            filename_or_obj,
            storage_options=storage_options,
            open_dataset_options=open_dataset_options,
        )
        return ref_ds

    open_dataset_parameters = [
        "filename_or_obj",
        "storage_options",
        "open_dataset_options",
    ]

    def guess_can_open(self, filename_or_obj):
        try:
            _, ext = os.path.splitext(filename_or_obj)
        except TypeError:
            return False
        return ext in {".json", ".json.zstd", ".parquet"}

    description = "Open Kerchunk References with Xarray"
    url = "https://fsspec.github.io/kerchunk/"


def open_reference_dataset(
    filename_or_obj, *, storage_options=None, open_dataset_options=None
):
    if storage_options is None:
        storage_options = {}
    if open_dataset_options is None:
        open_dataset_options = {}

    store = refs_as_store(filename_or_obj, **storage_options)

    return xr.open_zarr(
        store, zarr_format=2, consolidated=False, **open_dataset_options
    )



import fsspec
import fsspec.utils
import numpy as np
import dask.array as da
import pytest
import xarray as xr
import zarr.storage

import kerchunk.combine
from kerchunk.zarr import single_zarr
from kerchunk.combine import MultiZarrToZarr
from kerchunk.utils import fs_as_store, refs_as_store, consolidate


fs = fsspec.filesystem("memory")
arr = np.random.rand(1, 10, 10)


@pytest.fixture(scope="module", autouse=True)
def datasets():
    # if something fails here, it won't crash the test run
    static = xr.DataArray(data=np.random.rand(10, 10), dims=["x", "y"], name="static")
    data = xr.DataArray(
        data=arr.squeeze(),
        dims=["x", "y"],
        name="data",
    )
    xr.Dataset({"data": data}, attrs={"attr0": 3}).to_zarr(
        "memory://simple1.zarr", zarr_format=2
    )

    data = xr.DataArray(
        data=arr.squeeze() + 1,
        dims=["x", "y"],
        name="data",
    )
    xr.Dataset({"data": data}, attrs={"attr0": 4}).to_zarr(
        "memory://simple2.zarr", zarr_format=2
    )

    data = xr.DataArray(
        data=arr.squeeze(),
        dims=["x", "y"],
        name="datum",
    )
    xr.Dataset({"datum": data}, attrs={"attr0": 3}).to_zarr(
        "memory://simple_var1.zarr", zarr_format=2
    )

    data = xr.DataArray(
        data=arr.squeeze() + 1,
        dims=["x", "y"],
        name="datum",
    )
    xr.Dataset({"datum": data}, attrs={"attr0": 4}).to_zarr(
        "memory://simple_var2.zarr", zarr_format=2
    )

    data = xr.DataArray(
        data=arr,
        coords={"time": np.array([1])},
        dims=["time", "x", "y"],
        name="data",
        attrs={"attr0": 3},
    )
    xr.Dataset({"data": data, "static": static}, attrs={"attr1": 5}).to_zarr(
        "memory://single1.zarr", zarr_format=2
    )

    data = xr.DataArray(
        data=arr,
        coords={"time": np.array([2])},
        dims=["time", "x", "y"],
        name="data",
        attrs={"attr0": 4},
    )
    xr.Dataset({"data": data, "static": static}, attrs={"attr1": 6}).to_zarr(
        "memory://single2.zarr", zarr_format=2
    )

    data = xr.DataArray(
        data=arr,
        coords={"time": np.array([3])},
        dims=["time", "x", "y"],
        name="data",
        attrs={"attr0": 4},
    )
    xr.Dataset({"data": data, "static": static}, attrs={"attr1": 6}).to_zarr(
        "memory://single3.zarr", zarr_format=2
    )

    data = xr.DataArray(
        data=np.vstack([arr] * 4),
        coords={"time": np.array([1, 2, 3, 4])},
        dims=["time", "x", "y"],
        name="data",
        attrs={"attr0": 0},
    )
    xr.Dataset({"data": data}).to_zarr("memory://quad_nochunk1.zarr", zarr_format=2)
    xr.Dataset({"data": data}).to_zarr(
        "memory://group1.zarr", group="group", zarr_format=2
    )

    data = xr.DataArray(
        data=np.vstack([arr] * 4),
        coords={"time": np.array([5, 6, 7, 8])},
        dims=["time", "x", "y"],
        name="data",
        attrs={"attr0": 0},
    )
    xr.Dataset({"data": data}).to_zarr("memory://quad_nochunk2.zarr", zarr_format=2)
    xr.Dataset({"data": data}).to_zarr(
        "memory://group2.zarr", group="group", zarr_format=2
    )

    data = xr.DataArray(
        data=da.from_array(np.vstack([arr] * 4), chunks=(1, 10, 10)),
        coords={"time": np.array([1, 2, 3, 4])},
        dims=["time", "x", "y"],
        name="data",
        attrs={"attr0": 0},
    )
    xr.Dataset({"data": data}).to_zarr("memory://quad_1chunk1.zarr", zarr_format=2)

    data = xr.DataArray(
        data=da.from_array(np.vstack([arr] * 4), chunks=(1, 10, 10)),
        coords={"time": np.array([5, 6, 7, 8])},
        dims=["time", "x", "y"],
        name="data",
        attrs={"attr0": 0},
    )
    xr.Dataset({"data": data}).to_zarr("memory://quad_1chunk2.zarr", zarr_format=2)

    data = xr.DataArray(
        data=da.from_array(np.vstack([arr] * 4), chunks=(2, 10, 10)),
        coords={"time": np.array([1, 2, 3, 4])},
        dims=["time", "x", "y"],
        name="data",
        attrs={"attr0": 0},
    )
    xr.Dataset({"data": data}).to_zarr("memory://quad_2chunk1.zarr", zarr_format=2)

    data = xr.DataArray(
        data=da.from_array(np.vstack([arr] * 4), chunks=(2, 10, 10)),
        coords={"time": np.array([5, 6, 7, 8])},
        dims=["time", "x", "y"],
        name="data",
        attrs={"attr0": 0},
    )
    xr.Dataset({"data": data}).to_zarr("memory://quad_2chunk2.zarr", zarr_format=2)

    # simple time arrays - xarray can't make these!
    z = zarr.open("memory://time1.zarr", mode="w", zarr_format=2)
    time1_array = np.array([1], dtype="M8[s]")
    ar = z.create_array("time", shape=time1_array.shape, dtype=time1_array.dtype)
    ar[:] = time1_array
    ar.attrs.update({"_ARRAY_DIMENSIONS": ["time"]})
    ar = z.create_array("data", dtype=arr.dtype, shape=arr.shape)
    ar[:] = arr
    ar.attrs.update({"_ARRAY_DIMENSIONS": ["time", "x", "y"]})

    z = zarr.open("memory://time2.zarr", mode="w", zarr_format=2)
    time2_array = np.array([2], dtype="M8[s]")
    ar = z.create_array("time", dtype=time2_array.dtype, shape=time2_array.shape)
    ar[:] = time2_array
    ar.attrs.update({"_ARRAY_DIMENSIONS": ["time"]})
    ar = z.create_array("data", dtype=arr.dtype, shape=arr.shape)
    ar[:] = arr
    ar.attrs.update({"_ARRAY_DIMENSIONS": ["time", "x", "y"]})

    # cftime arrays - standard
    tdata1 = xr.DataArray(
        data=arr,
        coords={"time": np.array([1])},
        dims=["time", "x", "y"],
        name="data",
    )
    xr.Dataset({"data": tdata1}).to_zarr("memory://cfstdtime1.zarr", zarr_format=2)
    fs.pipe(
        "cfstdtime1.zarr/time/.zattrs",
        b'{"_ARRAY_DIMENSIONS": ["time"], "units": "seconds since '
        b'1970-01-01T00:00:00"}',
    )

    tdata1 = xr.DataArray(
        data=arr,
        coords={"time": np.array([2])},
        dims=["time", "x", "y"],
        name="data",
    )
    xr.Dataset({"data": tdata1}).to_zarr("memory://cfstdtime2.zarr", zarr_format=2)
    fs.pipe(
        "cfstdtime2.zarr/time/.zattrs",
        b'{"_ARRAY_DIMENSIONS": ["time"], "units": "seconds since '
        b'1970-01-01T00:00:00"}',
    )

    tdata1 = xr.DataArray(
        data=arr,
        coords={"time": np.array([3])},
        dims=["time", "x", "y"],
        name="data",
    )
    xr.Dataset({"data": tdata1}).to_zarr("memory://cfstdtime3.zarr", zarr_format=2)
    fs.pipe(
        "cfstdtime3.zarr/time/.zattrs",
        b'{"_ARRAY_DIMENSIONS": ["time"], "units": "seconds since '
        b'1970-01-01T00:00:00"}',
    )

    # cftime arrays - non standard
    tdata1 = xr.DataArray(
        data=arr,
        coords={"time": np.array([1])},
        dims=["time", "x", "y"],
        name="data",
        attrs={"units": "months since 1970-01-01", "calendar": "360_day"},
    )
    xr.Dataset({"data": tdata1}).to_zarr("memory://cfnontime1.zarr", zarr_format=2)
    fs.pipe(
        "cfnontime1.zarr/time/.zattrs",
        b'{"_ARRAY_DIMENSIONS": ["time"], "units": "months since 1970-01-01", "calendar": "360_day"}',
    )

    tdata1 = xr.DataArray(
        data=arr,
        coords={"time": np.array([2])},
        dims=["time", "x", "y"],
        name="data",
        attrs={"units": "months since 1970-01-01", "calendar": "360_day"},
    )
    xr.Dataset({"data": tdata1}).to_zarr("memory://cfnontime2.zarr", zarr_format=2)
    fs.pipe(
        "cfnontime2.zarr/time/.zattrs",
        b'{"_ARRAY_DIMENSIONS": ["time"], "units": "months since 1970-01-01", "calendar": "360_day"}',
    )


@pytest.fixture(scope="module")
def refs():
    return {
        path.replace(".zarr", "").lstrip("/"): single_zarr(f"memory://{path}")
        for path in fs.ls("", detail=False)
    }


def test_fixture(refs):
    # effectively checks that single_zarr works
    assert "single1" in refs
    fs = fsspec.filesystem("reference", fo=refs["single1"], remote_protocol="memory")
    store = zarr.storage.FsspecStore(fs)
    g = xr.open_dataset(store, engine="zarr", backend_kwargs={"consolidated": False})
    assert g.time.values.tolist() == [1]
    assert (g.data.values == arr).all()
    assert g.attrs["attr1"] == 5
    assert g.data.attrs["attr0"] == 3


@pytest.mark.parametrize(
    "dataset,chunks",
    [
        ["quad_nochunk1", ["data/0.0.0"]],
        ["quad_1chunk1", ["data/0.0.0", "data/1.0.0", "data/2.0.0", "data/3.0.0"]],
        ["quad_2chunk1", ["data/0.0.0", "data/1.0.0"]],
    ],
)
def test_fixture_chunks(refs, dataset, chunks):
    # checks that the right number of chunks are made per dataset
    fs = fsspec.filesystem("reference", fo=refs[dataset], remote_protocol="memory")
    out = fs.ls("data", detail=False)
    assert out == ["data/.zarray", "data/.zattrs"] + chunks
    out = fs.ls("time", detail=False)
    assert out == ["time/.zarray", "time/.zattrs", "time/0"]


@pytest.mark.parametrize(
    "selector,expected",
    [
        ["data:time", [1, 2]],
        [[9, 10], [9, 10]],
        ["INDEX", [0, 1]],
        # [re.compile("simple(\\d)"), [1, 2]],  # we don't give path names here
        ["attr:attr1", [5, 6]],
        ["vattr:data:attr0", [3, 4]],
    ],
)
def test_get_coos(refs, selector, expected):
    mzz = MultiZarrToZarr(
        [refs["single1"], refs["single2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_map={"time": selector},
    )
    mzz.first_pass()
    assert mzz.coos["time"].tolist() == expected
    mzz.store_coords()
    store = refs_as_store(mzz.out)
    g = zarr.open(store, zarr_format=2)
    assert g["time"][:].tolist() == expected
    assert dict(g.attrs)


def test_coo_vars(refs):
    mzz = MultiZarrToZarr(
        [refs["simple1"], refs["simple_var1"]],
        remote_protocol="memory",
        concat_dims=["var"],
    )
    mzz.first_pass()
    assert mzz.coos["var"].tolist() == ["data", "datum"]

    mzz = MultiZarrToZarr(
        [refs["simple1"], refs["simple2"], refs["simple_var1"], refs["simple_var2"]],
        remote_protocol="memory",
        concat_dims=["var", "time"],
        coo_map={"time": "attr:attr0"},
    )
    mzz.first_pass()
    assert mzz.coos["var"].tolist() == ["data", "datum"]
    assert mzz.coos["time"].tolist() == [3, 4]


def test_single(refs):
    mzz = MultiZarrToZarr(
        [refs["single1"], refs["single2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_dtypes={"time": "int16"},
    )
    mzz.first_pass()
    mzz.store_coords()
    mzz.second_pass()
    out = kerchunk.combine.consolidate(mzz.out)
    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {"fo": out, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
        decode_cf=False,
    )
    # TODO: make some assert_eq style function
    assert (
        z.time.dtype == "int16"
    )  # default is int64, or float64 if decode_cf is True because or array special name
    assert z.time.values.tolist() == [1, 2]
    assert z.data.shape == (2, 10, 10)
    assert (z.data[0].values == arr).all()
    assert (z.data[1].values == arr).all()


def test_single_append(refs):
    mzz = MultiZarrToZarr(
        [refs["single1"], refs["single2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_dtypes={"time": "int16"},
    )
    out = mzz.translate()
    mzz = MultiZarrToZarr.append(
        [refs["single3"]],
        out,
        remote_protocol="memory",
        concat_dims=["time"],
        coo_dtypes={"time": "int16"},
    )
    out = mzz.translate()
    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {"fo": out, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
        decode_cf=False,
    )
    assert z.data.shape == (3, 10, 10)
    assert out["refs"]["data/1.0.0"] == ["memory:///single2.zarr/data/0.0.0"]
    assert out["refs"]["data/2.0.0"] == ["memory:///single3.zarr/data/0.0.0"]
    assert z.time.values.tolist() == [1, 2, 3]


@pytest.mark.parametrize("mapper", [{}, {"time": "cf:time"}])
@pytest.mark.parametrize("dtype", [{"time": "M8[s]"}, {}])
def test_single_append_cf(refs, mapper, dtype):
    mzz = MultiZarrToZarr(
        [refs["cfstdtime1"], refs["cfstdtime2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_map=mapper,
        coo_dtypes=dtype,
    )
    out = mzz.translate()
    mzz = MultiZarrToZarr.append(
        [refs["cfstdtime3"]],
        out,
        remote_protocol="memory",
        concat_dims=["time"],
        coo_map=mapper,
        coo_dtypes=dtype,
    )
    out = mzz.translate()
    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {"fo": out, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
    )
    assert z.data.shape == (3, 10, 10)
    assert out["refs"]["data/0.0.0"] == ["memory:///cfstdtime1.zarr/data/0.0.0"]
    assert out["refs"]["data/1.0.0"] == ["memory:///cfstdtime2.zarr/data/0.0.0"]
    assert out["refs"]["data/2.0.0"] == ["memory:///cfstdtime3.zarr/data/0.0.0"]
    np.testing.assert_equal(
        z.time.values,
        np.array(
            [
                "1970-01-01T00:00:01.000000000",
                "1970-01-01T00:00:02.000000000",
                "1970-01-01T00:00:03.000000000",
            ],
            dtype="datetime64[ns]",
        ),
    )


def test_single_append_parquet(refs):
    from fsspec.implementations.reference import LazyReferenceMapper

    m = fsspec.filesystem("memory")
    out = LazyReferenceMapper.create("memory://refs/out.parquet", fs=m)
    mzz = MultiZarrToZarr(
        [refs["single1"], refs["single2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_dtypes={"time": "int16"},
        out=out,
    )
    mzz.translate()

    # reload here due to unknown bug after flush
    out = LazyReferenceMapper("memory://refs/out.parquet", fs=m)
    mzz = MultiZarrToZarr.append(
        [refs["single3"]],
        out,
        remote_protocol="memory",
        concat_dims=["time"],
        coo_dtypes={"time": "int16"},
    )
    out = mzz.translate()

    z = xr.open_dataset(
        out,
        backend_kwargs={
            "storage_options": {"remote_protocol": "memory"},
        },
        engine="kerchunk",
        decode_cf=False,
    )
    assert z.data.shape == (3, 10, 10)
    assert out["data/1.0.0"] == ["memory:///single2.zarr/data/0.0.0"]
    assert out["data/2.0.0"] == ["memory:///single3.zarr/data/0.0.0"]
    assert z.time.values.tolist() == [1, 2, 3]


def test_lazy_filler(tmpdir, refs):
    pd = pytest.importorskip("pandas")
    pytest.importorskip("fastparquet")
    from fsspec.implementations.reference import LazyReferenceMapper

    fs = fsspec.filesystem("file")
    tmpdir = str(tmpdir)
    out = LazyReferenceMapper.create(record_size=10, root=tmpdir, fs=fs)

    mzz = MultiZarrToZarr(
        [refs["single1"], refs["single2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_dtypes={"time": "int16"},
        out=out,
    )
    mzz.first_pass()
    mzz.store_coords()

    out.flush()
    allfiles = fs.find(tmpdir)
    assert allfiles == [f"{tmpdir}/{a}" for a in [".zmetadata", "time/refs.0.parq"]]
    df = pd.read_parquet(f"{tmpdir}/time/refs.0.parq", engine="fastparquet")
    assert df.iloc[:1].to_dict() == {
        "raw": {0: b"\x01\x00\x02\x00"},
        "offset": {0: 0},
        "size": {0: 0},
        "path": {0: None},
    }
    assert out["time/0"] == b"\x01\x00\x02\x00"
    mzz.second_pass()

    # actual references don't show
    assert set(out.zmetadata) == {
        ".zattrs",
        ".zgroup",
        "data/.zarray",
        "data/.zattrs",
        "static/.zarray",
        "static/.zattrs",
        "time/.zarray",
        "time/.zattrs",
    }
    assert out._items
    out.flush()
    assert set(out.zmetadata) == {
        ".zattrs",
        ".zgroup",
        "data/.zarray",
        "data/.zattrs",
        "static/.zarray",
        "static/.zattrs",
        "time/.zarray",
        "time/.zattrs",
    }
    assert set(out._items) == {".zmetadata"}
    allfiles = fs.find(tmpdir)
    assert [
        f"{tmpdir}/{a}" in allfiles for a in ["static/refs.0.parq", "data/refs.0.parq"]
    ]

    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {"fo": tmpdir, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
        decode_cf=False,
    )
    assert z.time.values.tolist() == [1, 2]
    assert z.data.shape == (2, 10, 10)
    assert (z.data[0].values == arr).all()
    assert (z.data[1].values == arr).all()


def test_run_twice(refs):
    mzz = MultiZarrToZarr(
        [refs["single1"], refs["single2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_dtypes={"time": "int16"},
    )
    mzz.translate()
    out = mzz.translate()
    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {"fo": out, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
        decode_cf=False,
    )
    assert (
        z.time.dtype == "int16"
    )  # default is int64, or float64 if decode_cf is True because or array special name
    assert z.time.values.tolist() == [1, 2]
    assert z.data.shape == (2, 10, 10)
    assert (z.data[0].values == arr).all()
    assert (z.data[1].values == arr).all()


def test_outfile_postprocess(refs):
    def post_process(ref):
        # renamed "data" array to "a_data"; to rename a coordinate, one would nee
        # to alter attributes of arrays pointing to it
        return {("a_" + k if k.startswith("data") else k): v for k, v in ref.items()}

    mzz = MultiZarrToZarr(
        [refs["single1"], refs["single2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        postprocess=post_process,
    )
    mzz.translate("memory://myout.json")
    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {
                "fo": "memory://myout.json",
                "remote_protocol": "memory",
            },
            "consolidated": False,
        },
        engine="zarr",
    )
    assert z.time.values.tolist() == [1, 2]
    assert z.a_data.shape == (2, 10, 10)
    assert (z.a_data[0].values == arr).all()
    assert (z.a_data[1].values == arr).all()


@pytest.mark.parametrize(
    "inputs, chunks",
    [
        [["quad_nochunk1", "quad_nochunk2"], ((4, 4), (10,), (10,))],
        [["quad_1chunk1", "quad_1chunk2"], ((1,) * 8, (10,), (10,))],
        [["quad_2chunk1", "quad_2chunk2"], ((2,) * 4, (10,), (10,))],
        [["group1", "group2"], ((4, 4), (10,), (10,))],
    ],
)
def test_chunked(refs, inputs, chunks):
    mzz = MultiZarrToZarr(
        [refs[inputs[0]], refs[inputs[1]]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_map={"time": "data:group/time"} if "group" in inputs[0] else None,
    )
    out = mzz.translate()
    z = xr.open_dataset(
        f"reference://{'group' if 'group' in inputs[0] else ''}",
        backend_kwargs={
            "storage_options": {"fo": out, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
        chunks={},
    )
    # TODO: make some assert_eq style function
    assert z["time"].values.tolist() == [1, 2, 3, 4, 5, 6, 7, 8]
    assert z["data"].shape == (8, 10, 10)
    assert z["data"].chunks == chunks
    for i in range(z.data.shape[0]):
        assert (z.data[i].values == arr).all()


def test_var(refs):
    mzz = MultiZarrToZarr(
        [refs["simple1"], refs["simple_var1"]],
        remote_protocol="memory",
        concat_dims=["var"],
    )
    out = mzz.translate()
    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {"fo": out, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
        chunks={},
    )
    assert list(z) == ["data", "datum"]
    assert list(z.dims) == ["x", "y"]
    assert z.data.shape == z.datum.shape == (10, 10)
    assert (z.data.values == arr).all()
    assert (z.datum.values == arr).all()


def test_var_dimension(refs):
    mzz = MultiZarrToZarr(
        [refs["simple1"], refs["simple_var1"]],
        remote_protocol="memory",
        coo_map={"varname": "VARNAME", "var": "output"},
    )
    out = mzz.translate()
    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {"fo": out, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
        chunks={},
    )
    assert list(z) == ["output"]
    assert list(z.dims) == ["varname", "x", "y"]
    assert list(z.varname) == ["data", "datum"]
    assert z.output.shape == (2, 10, 10)
    assert (z.output.values == arr).all()


def test_var_and_dim(refs):
    mzz = MultiZarrToZarr(
        [refs["simple1"], refs["simple2"], refs["simple_var1"], refs["simple_var2"]],
        remote_protocol="memory",
        concat_dims=["var", "dim"],
        coo_map={"dim": "attr:attr0"},
    )
    out = mzz.translate()
    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {"fo": out, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
        chunks={},
    )
    assert list(z) == ["data", "datum"]
    assert list(z.dims) == ["dim", "x", "y"]
    assert z.data.shape == z.datum.shape == (2, 10, 10)
    assert z["dim"].values.tolist() == [3, 4]
    assert (z.data.values == np.vstack([arr, arr + 1])).all()


@pytest.mark.parametrize(
    "inps,expected",
    [
        ["time", ["1970-01-01T00:00:01", "1970-01-01T00:00:02"]],
        ["cfstdtime", ["1970-01-01T00:00:01", "1970-01-01T00:00:02"]],
        ["cfnontime", ["1970-02-01T00:00:00", "1970-03-01T00:00:00"]],
    ],
)
def test_times(refs, inps, expected):
    mapper = "data:time" if inps == "time" else "cf:time"
    mzz = MultiZarrToZarr(
        [refs[inps + "1"], refs[inps + "2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_map={"time": mapper},
    )
    out = mzz.translate()
    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {"fo": out, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
        chunks={},
    )
    if inps == "cfnontime":
        assert [_.isoformat() for _ in z.time.values] == expected
    else:
        assert (z.time == np.array(expected, dtype="M8[ms]")).all()


def test_cftimes_to_normal(refs):
    mzz = MultiZarrToZarr(
        [refs["cfnontime1"], refs["cfnontime2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_map={"time": "cf:time"},
        coo_dtypes={"time": "M8[s]"},
    )
    out = mzz.translate()
    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {"fo": out, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
        chunks={},
    )
    assert z.time.dtype.kind == "M"
    assert (
        z.time.values
        == np.array(["1970-02-01T00:00:00", "1970-03-01T00:00:00"], dtype="M8[s]")
    ).all()


def test_inline(refs):
    mzz = MultiZarrToZarr(
        [refs["single1"], refs["single2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_dtypes={"time": "int16"},
        inline_threshold=50000,
    )
    out = mzz.translate()
    z = xr.open_dataset(
        "reference://",
        backend_kwargs={
            "storage_options": {"fo": out, "remote_protocol": "memory"},
            "consolidated": False,
        },
        engine="zarr",
        decode_cf=False,
    )
    ref = fsspec.filesystem("reference", fo=out, remote_protocol="memory")
    assert (
        z.time.dtype == "int16"
    )  # default is int64, or float64 if decode_cf is True because or array special name
    assert z.time.values.tolist() == [1, 2]
    assert z.data.shape == (2, 10, 10)
    assert (z.data[0].values == arr).all()
    assert (z.data[1].values == arr).all()

    assert isinstance(ref.references["data/0.0.0"], str)
    assert ref.references["data/0.0.0"].startswith("base64:")


def test_no_inline():
    """Ensure that inline_threshold=0 disables MultiZarrToZarr checking file size."""
    ds = xr.Dataset(dict(x=[1, 2, 3]))
    ds["y"] = 3 + ds["x"]
    ds.to_zarr("memory://zarr_store", mode="w", zarr_format=2, consolidated=False)
    store = fsspec.get_mapper("memory://zarr_store")
    ref = kerchunk.utils.consolidate(store)
    # This type of reference with no offset or total size is produced by
    # kerchunk.zarr.single_zarr or equivalently ZarrToZarr.translate.
    ref["refs"]["y/0"] = ["file:///tmp/some/data-that-shouldnt-be-accessed"]

    mzz_no_inline = MultiZarrToZarr([ref], concat_dims=["x"], inline_threshold=0)
    # Should be okay because inline_threshold=None so we don't check the file size
    # in order to see if it should be inlined
    mzz_no_inline.translate()

    mzz_inline = MultiZarrToZarr([ref], concat_dims=["x"], inline_threshold=1)
    with pytest.raises(FileNotFoundError):
        # Should raise because we check the file size to see if it should be inlined,
        # and the example was engineered so that the file doesn't exist.
        mzz_inline.translate()


def test_merge_vars():
    a = dict({"version": 1, "refs": dict({"item1": 1})})
    b = dict({"version": 1, "refs": dict({"item2": 2})})
    merge = kerchunk.combine.merge_vars([a, b])
    assert list(merge["refs"]) == ["item1", "item2"]
    fs = fsspec.filesystem("memory")
    fs.pipe("file1.json", b"""{"version": 1, "refs": {"item1": 1}}""")
    fs.pipe("file2.json", b"""{"version": 1, "refs": {"item2": 2}}""")
    merge = kerchunk.combine.merge_vars(["memory://file1.json", "memory://file2.json"])
    assert list(merge["refs"]) == ["item1", "item2"]


def test_bad_coo_warning(refs):
    def f(*_, **__):
        return 1

    mzz = MultiZarrToZarr(
        [refs["single1"], refs["single2"]],
        remote_protocol="memory",
        concat_dims=["time"],
        coo_map={"time": f},
    )
    with pytest.warns(match="contains less than expected"):
        mzz.first_pass()


def test_chunk_error(refs):
    # Work with the refs in the output of single_zarr
    refs1 = refs["single1"]["refs"]
    refs2 = refs1.copy()
    refs2.pop(".zmetadata")
    fs = fsspec.filesystem("reference", fo=refs2, remote_protocol="memory")  # noqa
    refs2[
        "data/.zarray"
    ] = b"""
    {
    "chunks": [2, 3, 3],
    "zarr_format": 2,
    "shape": [1, 10, 10]
    }
    """
    mzz = MultiZarrToZarr(
        [refs1, refs2],
        remote_protocol="memory",
        concat_dims=["time"],
    )
    with pytest.raises(ValueError) as e:
        mzz.translate()
    assert "chunk size mismatch" in str(e.value)



import fsspec
import numpy as np
import ujson

from kerchunk.utils import refs_as_store


decoders = {}


def reg(name):
    def f(func):
        decoders[name] = func
        return func

    return f


class HDF4ToZarr:
    """Experimental: interface to HDF4 archival files"""

    def __init__(
        self,
        path,
        storage_options=None,
        inline_threshold=100,
        out=None,
    ):
        self.path = path
        self.st = storage_options
        self.thresh = inline_threshold
        self.out = out or {}

    def read_int(self, n):
        return int.from_bytes(self.f.read(n), "big")

    def read_ddh(self):
        return {"ndd": self.read_int(2), "next": self.read_int(4)}

    def read_dd(self):
        loc = self.f.tell()
        i = int.from_bytes(self.f.read(2), "big")
        if i & 0x4000:
            extended = True
            i = i - 0x4000
        else:
            extended = False
        tag = tags.get(i, i)
        no_data = tag not in {"NULL"}
        ref = (tag, int.from_bytes(self.f.read(2), "big"))
        info = {
            "offset": int.from_bytes(self.f.read(4), "big") * no_data,
            "length": int.from_bytes(self.f.read(4), "big") * no_data,
            "extended": extended,
            "loc": loc,
        }
        return ref, info

    def decode(self, tag, info):
        self.f.seek(info["offset"])
        ident = lambda _, __: info
        return decoders.get(tag, ident)(self, info)

    def translate(self, filename=None, storage_options=None):
        """Scan and return references

        Parameters
        ----------
        filename: if given, write to this as JSON
        storage_options: to interpret filename

        Returns
        -------
        references
        """
        import zarr
        from kerchunk.codecs import ZlibCodec

        fo = fsspec.open(self.path, **(self.st or {}))
        self.f = fo.open()

        # magic header
        assert self.f.read(4) == b"\x0e\x03\x13\x01"

        # all the data descriptors in a linked list
        self.tags = {}
        while True:
            ddh = self.read_ddh()

            for _ in range(ddh["ndd"]):
                ident, info = self.read_dd()
                self.tags[ident] = info
            if ddh["next"] == 0:
                # "finished" sentry
                break
            # or continue
            self.f.seek(ddh["next"])

        # basic decode
        for tag, ref in self.tags:
            self._dec(tag, ref)

        # global attributes
        attrs = {}
        for (tag, ref), info in self.tags.items():
            if tag == "VH" and info["names"][0].upper() == "VALUES":
                # dtype = dtypes[info["types"][0]]
                inf2 = self.tags[("VS", ref)]
                self.f.seek(inf2["offset"])
                # remove zero padding
                data = self.f.read(inf2["length"]).split(b"\x00", 1)[0]
                # NASA conventions
                if info["name"].startswith(
                    ("CoreMetadata.", "ArchiveMetadata.", "StructMetadata.")
                ):
                    obj = None
                    for line in data.decode().split("\n"):
                        if "OBJECT" in line:
                            obj = line.split()[-1]
                        if "VALUE" in line:
                            attrs[obj] = line.split()[-1].lstrip('"').rstrip('"')

        # there should be only one root, and it's probably the last VG
        # so maybe this loop isn't needed
        roots = set()
        children = set()
        child = {}
        for (tag, ref), info in self.tags.items():
            if tag == "VG":
                here = child.setdefault((tag, ref), set())
                for t, r in zip(info["tag"], info["refs"]):
                    if t == "VG":
                        children.add((t, r))
                        roots.discard((t, r))
                        here.add((t, r))
                if tag not in children:
                    roots.add((tag, ref))

        # hierarchical output
        output = self._descend_vg(*sorted(roots, key=lambda t: t[1])[-1])
        prot = fo.fs.protocol
        prot = prot[0] if isinstance(prot, tuple) else prot
        store = refs_as_store(self.out, remote_protocol=prot, remote_options=self.st)
        g = zarr.open_group(store, zarr_format=2, use_consolidated=False)
        refs = {}
        for k, v in output.items():
            if isinstance(v, dict):
                compressor = ZlibCodec() if "refs" in v else None
                arr = g.require_array(
                    name=k,
                    shape=v["dims"],
                    dtype=v["dtype"],
                    chunks=v.get("chunks", v["dims"]),
                    compressor=compressor,
                )
                arr.attrs.update(
                    dict(
                        _ARRAY_DIMENSIONS=(
                            [f"{k}_x", f"{k}_y"][: len(v["dims"])]
                            if "refs" in v
                            else ["0"]
                        ),
                        **{
                            i: j.tolist() if isinstance(j, np.generic) else j
                            for i, j in v.items()
                            if i not in {"chunk", "dims", "dtype", "refs"}
                        },
                    )
                )
                for r in v.get("refs", []):
                    if r[0] == "DEFLATE":
                        continue
                    refs[f"{k}/{r[0]}"] = [self.path, r[1], r[2]]
            else:
                if not k.startswith(
                    ("CoreMetadata.", "ArchiveMetadata.", "StructMetadata.")
                ):
                    attrs[k] = v.tolist() if isinstance(v, np.generic) else v
        store.fs.references.update(refs)
        g.attrs.update(attrs)

        if filename is None:
            return store.fs.references
        with fsspec.open(filename, **(storage_options or {})) as f:
            ujson.dumps(dict(store.fs.references), f)

    def _descend_vg(self, tag, ref):
        info = self.tags[(tag, ref)]
        out = {}
        for t, r in zip(info["tag"], info["refs"]):
            inf2 = self.tags[(t, r)]
            if t == "VG":
                tmp = self._descend_vg(t, r)
                if tmp and list(tmp)[0] == inf2["name"]:
                    tmp = tmp[inf2["name"]]
                out[inf2["name"]] = tmp
            elif t == "VH":
                if len(inf2["names"]) == 1 and inf2["names"][0].lower() == "values":
                    dtype = dtypes[inf2["types"][0]]
                    name = inf2["name"]
                    inf2 = self.tags[("VS", r)]
                    self.f.seek(inf2["offset"])
                    data = self.f.read(inf2["length"])
                    if dtype == "str":
                        out[name] = (
                            data.split(b"\x00", 1)[0].decode().lstrip('"').rstrip('"')
                        )  # decode() ?
                    else:
                        out[name] = np.frombuffer(data, dtype)[0]
            elif t == "NT":
                out["dtype"] = inf2["typ"]
            elif t == "SD":
                if isinstance(inf2["data"][-1], (tuple, list)):
                    out["refs"] = inf2["data"][:-1]
                    out["chunks"] = [_["chunk_length"] for _ in inf2["data"][-1]]
                else:
                    out["refs"] = [inf2["data"]]
                    out["chunks"] = True
            elif t == "SDD":
                out["dims"] = inf2["dims"]
            elif t == "NDG":
                pass  # out.setdefault("extra", []).append(_dec_ndg(self, inf2))
        if out.get("chunks") is True:
            out["chunks"] = out["dims"]
            out["refs"] = [
                [".".join(["0"] * len(out["dims"]))]
                + [out["refs"][0][1], out["refs"][0][2], out["refs"][0][0]]
            ]
        return out

    def _dec(self, tag, ref):
        info = self.tags[(tag, ref)]
        if not set(info) - {"length", "offset", "extended", "loc"}:
            self.f.seek(info["offset"])
            if info["extended"]:
                info["data"] = self._dec_extended()
            else:
                info.update(self.decode(tag, info))
        return info

    def _dec_extended(self):
        ext_type = spec[self.read_int(2)]
        if ext_type == "CHUNKED":
            return self._dec_chunked()
        elif ext_type == "LINKED":
            return self._dec_linked_header()
        elif ext_type == "COMP":
            return self._dec_comp()

    def _dec_linked_header(self):
        # get the bytes of a linked set - these will always be inlined
        self.read_int(4)  # length
        self.read_int(4)  # blk_len
        self.read_int(4)  # num_blk
        next_ref = self.read_int(2)
        out = []
        while next_ref:
            next_ref, data = self._dec_linked_block(self.tags[("LINKED", next_ref)])
            out.extend([d for d in data if d])
        bits = []
        for ref in out:
            info = self.tags[("LINKED", ref)]
            self.f.seek(info["offset"])
            bits.append(self.f.read(info["length"]))
        return b"".join(bits)

    def _dec_linked_block(self, block):
        self.f.seek(block["offset"])
        next_ref = self.read_int(2)
        refs = [self.read_int(2) for _ in range((block["length"] // 2) - 1)]
        return next_ref, refs

    def _dec_chunked(self):
        # we want to turn the chunks table into references
        # tag_head_len = self.read_int(4)
        # version = self.f.read(1)[0]
        # flag = self.read_int(4)
        # elem_tot_len = self.read_int(4)
        # chunk_size = self.read_int(4)
        # nt_size = self.read_int(4)
        self.f.seek(21, 1)
        chk_tbl_tag = tags[self.read_int(2)]  # should be VH
        chk_tbl_ref = self.read_int(2)
        self.read_int(2)  # sp_tab = tags[self.read_int(2)]
        self.read_int(2)  # sp_ref
        ndims = self.read_int(4)

        dims = [  # we don't use these, could skip
            {
                "flag": self.read_int(4),
                "dim_length": self.read_int(4),
                "chunk_length": self.read_int(4),
            }
            for _ in range(ndims)
        ]
        self.f.read(  # fill_value
            self.read_int(4)
        )  # to be interpreted as a number later; but chunk table probs has no fill
        # self.f.seek(12*ndims + 4, 1)  # if skipping

        header = self._dec(chk_tbl_tag, chk_tbl_ref)
        data = self._dec("VS", chk_tbl_ref)["data"]  # corresponding table

        # header gives the field pattern for the rows of data, one per chunk
        # maybe faster to use struct and iter than numpy, since we iterate anyway
        dt = [(f"ind{i}", ">u4") for i in range(ndims)] + [
            ("tag", ">u2"),
            ("ref", ">u2"),
        ]
        rows = np.frombuffer(data, dtype=dt, count=header["nvert"])
        # rows["tag"] should always be 61 -> CHUNK
        refs = []
        for *ind, tag, ref in rows:
            # maybe ind needs reversing since everything is FORTRAN
            chunk_tag = self.tags[("CHUNK", ref)]
            if chunk_tag["extended"]:
                self.f.seek(chunk_tag["offset"])
                # these are always COMP?
                ctype, offset, length = self._dec_extended()
                refs.append([".".join(str(_) for _ in ind), offset, length, ctype])
            else:
                refs.append(
                    [
                        ".".join(str(_) for _ in ind),
                        chunk_tag["offset"],
                        chunk_tag["length"],
                    ]
                )
        refs.append(dims)
        return refs

    def _dec_comp(self):
        # version = self.read_int(2)  # always 0
        # len_uncomp = self.read_int(4)
        self.f.seek(6, 1)

        data_ref = self.read_int(2)
        # model = self.read_int(2)  # always 0
        ctype = "DEFLATE"  # comp[self.read_int(2)]
        tag = self.tags[("COMPRESSED", data_ref)]
        return ctype, tag["offset"], tag["length"]


@reg("NDG")
def _dec_ndg(self, info):
    if "tags" not in info:
        return {
            "tags": [
                (tags[self.read_int(2)], self.read_int(2))
                for _ in range(0, info["length"], 4)
            ]
        }
    return info["tags"]


@reg("SDD")
def _dec_sdd(self, info):
    rank = self.read_int(2)
    dims = [self.read_int(4) for _ in range(rank)]
    data_tag = (tags[self.read_int(2)], self.read_int(2))
    scale_tags = [(tags[self.read_int(2)], self.read_int(2)) for _ in range(rank)]
    return _pl(locals())


@reg("VERSION")
def _dec_version(self, info):
    return {
        "major": self.read_int(4),
        "minor": self.read_int(4),
        "release": self.read_int(4),
        "string:": _null_str(self.f.read(info["length"] - 10).decode()),
    }


@reg("VH")
def _dec_vh(self, info):
    # virtual group ("table") header
    interface = self.read_int(2)
    nvert = self.read_int(4)
    ivsize = self.read_int(2)
    nfields = self.read_int(2)
    types = [self.read_int(2) for _ in range(nfields)]
    isize = [self.read_int(2) for _ in range(nfields)]
    offsets = [self.read_int(2) for _ in range(nfields)]
    order = [self.read_int(2) for _ in range(nfields)]
    names = [
        self.f.read(self.read_int(2)).split(b"\x00", 1)[0].decode()
        for _ in range(nfields)
    ]
    namelen = self.read_int(2)
    name = self.f.read(namelen).split(b"\x00", 1)[0].decode()
    classlen = self.read_int(2)
    cls = self.f.read(classlen).split(b"\x00", 1)[0].decode()
    ref = (self.read_int(2), self.read_int(2))
    return _pl(locals())


@reg("VG")
def _dec_vg(self, info):
    nelt = self.read_int(2)
    tag = [tags[self.read_int(2)] for _ in range(nelt)]
    refs = [self.read_int(2) for _ in range(nelt)]
    name = self.f.read(self.read_int(2)).split(b"\x00", 1)[0].decode()
    cls = self.f.read(self.read_int(2)).split(b"\x00", 1)[0].decode()
    return _pl(locals())


@reg("NT")
def _dec_nt(self, info):
    version, typ, width, cls = list(self.f.read(4))
    typ = dtypes[typ]
    return _pl(locals())


def _null_str(s):
    return s.split("\00", 1)[0]


def _pl(l):
    return {k: v for k, v in l.items() if k not in {"info", "f", "self"}}


# hdf/src/htags.h
tags = {
    1: "NULL",
    20: "LINKED",
    30: "VERSION",
    40: "COMPRESSED",
    50: "VLINKED",
    51: "VLINKED_DATA",
    60: "CHUNKED",
    61: "CHUNK",
    100: "FID",
    101: "FD",
    102: "TID",
    103: "TD",
    104: "DIL",
    105: "DIA",
    106: "NT",
    107: "MT",
    108: "FREE",
    200: "ID8",
    201: "IP8",
    202: "RI8",
    203: "CI8",
    204: "II8",
    300: "ID",
    301: "LUT",
    302: "RI",
    303: "CI",
    304: "NRI",
    306: "RIG",
    307: "LD",
    308: "MD",
    309: "MA",
    310: "CCN",
    311: "CFM",
    312: "AR",
    400: "DRAW",
    401: "RUN",
    500: "XYP",
    501: "MTO",
    602: "T14",
    603: "T105",
    700: "SDG",
    701: "SDD",
    702: "SD",
    703: "SDS",
    704: "SDL",
    705: "SDU",
    706: "SDF",
    707: "SDM",
    708: "SDC",
    709: "SDT",
    710: "SDLNK",
    720: "NDG",
    721: "RESERVED",
    # "Objects of tag 721 are never actually written to the file. The tag is
    # needed to make things easier mixing DFSD and SD style objects in the same file"
    731: "CAL",
    732: "FV",
    799: "BREQ",
    781: "SDRAG",
    780: "EREQ",
    1965: "VG",
    1962: "VH",
    1963: "VS",
    11: "RLE",
    12: "IMCOMP",
    13: "JPEG",
    14: "GREYJPEG",
    15: "JPEG5",
    16: "GREYJPEG5",
}
spec = {
    1: "LINKED",
    2: "EXT",
    3: "COMP",
    4: "VLINKED",
    5: "CHUNKED",
    6: "BUFFERED",
    7: "COMPRAS",
}

# hdf4/hdf/src/hntdefs.h
dtypes = {
    5: "f4",
    6: "f8",
    20: "i1",
    21: "u1",
    4: "str",  # special case, size given in header
    22: ">i2",
    23: ">u2",
    24: ">i4",
    25: ">u4",
    26: ">i8",
    27: ">u8",
}

# hdf4/hdf/src/hcomp.h
comp = {
    0: "NONE",
    1: "RLE",
    2: "NBIT",
    3: "SKPHUFF",
    4: "DEFLATE",  # called deflate, but code says "gzip" and doc says "GNU zip"; actually zlib?
    # see codecs.ZlibCodec
    5: "SZIP",
    7: "JPEG",
}
